from __future__ import annotations

# standard libraries
import functools
import gettext
import operator
import random
import string
import sys
import typing
import uuid
import weakref

# third party libraries
# None

# local libraries
from nion.swift import DataItemThumbnailWidget
from nion.swift import EntityBrowser
from nion.swift import Inspector
from nion.swift import MimeTypes
from nion.swift import NotificationDialog
from nion.swift import Undo
from nion.swift.model import Changes
from nion.swift.model import DataItem
from nion.swift.model import DisplayItem
from nion.swift.model import DocumentModel
from nion.swift.model import Notification
from nion.swift.model import Persistence
from nion.swift.model import Symbolic
from nion.ui import CanvasItem
from nion.ui import Declarative
from nion.ui import Dialog
from nion.ui import DrawingContext
from nion.ui import UserInterface
from nion.utils import Binding
from nion.utils import Converter
from nion.utils import Event
from nion.utils import Geometry
from nion.utils import ListModel
from nion.utils import Model
from nion.utils import Observable
from nion.utils import ReferenceCounting
from nion.utils import Registry

if typing.TYPE_CHECKING:
    from nion.swift import DocumentController
    from nion.ui import Application

_DocumentControllerWeakRefType = typing.Callable[[], "DocumentController.DocumentController"]

_ = gettext.gettext


class AddVariableCommand(Undo.UndoableCommand):

    def __init__(self, document_model: DocumentModel.DocumentModel, computation: Symbolic.Computation,
                 name: typing.Optional[str] = None, value_type: typing.Optional[Symbolic.ComputationVariableType] = None, value: typing.Any = None,
                 value_default: typing.Any = None, value_min: typing.Any = None, value_max: typing.Any = None,
                 control_type: typing.Optional[str] = None,
                 specified_item: typing.Optional[Persistence.PersistentObject] = None,
                 label: typing.Optional[str] = None):
        super().__init__(_("Add Computation Variable"))
        self.__document_model = document_model
        self.__computation_proxy = computation.create_proxy()
        self.__name = name
        self.__value_type = value_type
        self.__value = value
        self.__value_default = value_default
        self.__value_min = value_min
        self.__value_max = value_max
        self.__control_type = control_type
        self.__specified_item = specified_item
        self.__label = label
        self.__variable_index = 0
        self.initialize()

    def close(self) -> None:
        self.__computation_proxy.close()
        self.__computation_proxy = typing.cast(typing.Any, None)
        self.__name = None
        self.__value_type = None
        self.__value = None
        self.__value_default = None
        self.__value_min = None
        self.__value_max = None
        self.__control_type = None
        self.__specified_item = None
        self.__label = None
        self.__variable = None
        super().close()

    def _perform(self) -> None:
        computation = self.__computation_proxy.item
        if computation:
            variable = computation.create_variable(self.__name, self.__value_type, self.__value, self.__value_default, self.__value_min, self.__value_max, self.__control_type, self.__specified_item, self.__label)
            self.__variable_index = computation.variables.index(variable)

    def _get_modified_state(self) -> typing.Any:
        computation = self.__computation_proxy.item
        computation_modified_state = computation.modified_state if computation else None
        return computation_modified_state, self.__document_model.modified_state

    def _set_modified_state(self, modified_state: typing.Any) -> None:
        computation = self.__computation_proxy.item
        if computation:
            computation.modified_state = modified_state[0]
        self.__document_model.modified_state = modified_state[1]

    def _compare_modified_states(self, state1: typing.Any, state2: typing.Any) -> bool:
        # override to allow the undo command to track state; but only use part of the state for comparison
        return bool(state1[0] == state2[0])

    def _undo(self) -> None:
        computation = self.__computation_proxy.item
        if computation:
            variable = computation.variables[self.__variable_index]
            computation.remove_variable(variable)

    def _redo(self) -> None:
        self.perform()


class RemoveVariableCommand(Undo.UndoableCommand):

    def __init__(self, document_model: DocumentModel.DocumentModel, computation: Symbolic.Computation, variable: Symbolic.ComputationVariable) -> None:
        super().__init__(_("Remove Computation Variable"))
        self.__document_model = document_model
        self.__computation_proxy = computation.create_proxy()
        self.__variable_index = computation.variables.index(variable)
        self.__variable_dict = variable.write_to_dict()
        self.initialize()

    def close(self) -> None:
        self.__computation_proxy.close()
        self.__computation_proxy = typing.cast(typing.Any, None)
        self.__variable_dict = typing.cast(typing.Any, None)
        super().close()

    def _perform(self) -> None:
        computation = self.__computation_proxy.item
        if computation:
            variable = computation.variables[self.__variable_index]
            computation.remove_variable(variable)

    def _get_modified_state(self) -> typing.Any:
        computation = self.__computation_proxy.item
        computation_modified_state = computation.modified_state if computation else None
        return computation_modified_state, self.__document_model.modified_state

    def _set_modified_state(self, modified_state: typing.Any) -> None:
        computation = self.__computation_proxy.item
        if computation:
            computation.modified_state = modified_state[0]
        self.__document_model.modified_state = modified_state[1]

    def _compare_modified_states(self, state1: typing.Any, state2: typing.Any) -> bool:
        # override to allow the undo command to track state; but only use part of the state for comparison
        return bool(state1[0] == state2[0])

    def _undo(self) -> None:
        computation = self.__computation_proxy.item
        if computation:
            variable = Symbolic.ComputationVariable()
            variable.begin_reading()
            variable.read_from_dict(self.__variable_dict)
            variable.finish_reading()
            computation.insert_variable(self.__variable_index, variable)

    def _redo(self) -> None:
        computation = self.__computation_proxy.item
        if computation:
            computation.remove_variable(computation.variables[self.__variable_index])


class CreateComputationCommand(Undo.UndoableCommand):

    def __init__(self, document_model: DocumentModel.DocumentModel, data_item: DataItem.DataItem) -> None:
        super().__init__(_("Create Computation"))
        self.__document_model = document_model
        self.__data_item_proxy = data_item.create_proxy()

    def close(self) -> None:
        self.__document_model = typing.cast(typing.Any, None)
        self.__data_item_proxy.close()
        self.__data_item_proxy = typing.cast(typing.Any, None)
        super().close()

    @property
    def _computation(self) -> Symbolic.Computation:
        return self.__document_model.create_computation()

    def _perform(self) -> None:
        data_item = self.__data_item_proxy.item
        if data_item:
            computation = self.__document_model.create_computation()
            self.__document_model.set_data_item_computation(data_item, computation)

    def _get_modified_state(self) -> typing.Any:
        return self.__document_model.modified_state

    def _set_modified_state(self, modified_state: typing.Any) -> None:
        self.__document_model.modified_state = modified_state

    def _undo(self) -> None:
        data_item = self.__data_item_proxy.item
        if data_item:
            self.__document_model.set_data_item_computation(data_item, None)


class ChangeComputationCommand(Undo.UndoableCommand):

    def __init__(self, document_model: DocumentModel.DocumentModel, computation: Symbolic.Computation, *,
                 title: typing.Optional[str] = None, command_id: typing.Optional[str] = None,
                 is_mergeable: bool = False, **kwargs: typing.Any) -> None:
        super().__init__(title if title else _("Change Computation"), command_id=command_id, is_mergeable=is_mergeable)
        self.__document_model = document_model
        self.__computation_proxy = computation.create_proxy()
        self.__properties = {key: getattr(computation, key) for key in kwargs.keys()}
        self.__value_dict = kwargs
        self.initialize()

    def close(self) -> None:
        self.__properties = dict()
        self.__computation_proxy.close()
        self.__computation_proxy = typing.cast(typing.Any, None)
        self.__value_dict = typing.cast(typing.Any, None)
        super().close()

    def _perform(self) -> None:
        computation = self.__computation_proxy.item
        if computation:
            for key, value in self.__value_dict.items():
                setattr(computation, key, value)

    def _get_modified_state(self) -> typing.Any:
        computation = self.__computation_proxy.item
        computation_modified_state = computation.modified_state if computation else None
        return computation_modified_state, self.__document_model.modified_state

    def _set_modified_state(self, modified_state: typing.Any) -> None:
        computation = self.__computation_proxy.item
        if computation:
            computation.modified_state = modified_state[0]
        self.__document_model.modified_state = modified_state[1]

    def _compare_modified_states(self, state1: typing.Any, state2: typing.Any) -> bool:
        # override to allow the undo command to track state; but only use part of the state for comparison
        return bool(state1[0] == state2[0])

    def _undo(self) -> None:
        computation = self.__computation_proxy.item
        if computation:
            properties = self.__properties
            self.__properties = computation.write_to_dict()
            # NOTE: use read_properties_from_dict (read properties only), not read_from_dict (used for initialization).
            computation.read_properties_from_dict(properties)

    def can_merge(self, command: Undo.UndoableCommand) -> bool:
        return isinstance(command, ChangeComputationCommand) and bool(self.command_id) and self.command_id == command.command_id and self.__computation_proxy.item == command.__computation_proxy.item


def select_computation(document_model: DocumentModel.DocumentModel, display_item: typing.Optional[DisplayItem.DisplayItem]) -> typing.Optional[Symbolic.Computation]:
    if display_item:
        match_items: typing.Set[Persistence.PersistentObject] = set()
        match_items.add(display_item)
        match_items.update(display_item.data_items)
        match_items.update(display_item.graphics)
        computations: typing.Set[Symbolic.Computation] = set()
        for computation in document_model.computations:
            if set(computation.output_items).intersection(match_items):
                computations.add(computation)
        return next(iter(computations)) if len(computations) == 1 else None
    return None


class ComputationModel:
    """Represents a computation. Tracks a computation for changes to it and its content.

    Provides read/write access to the computation_text via the property.

    Provides a computation_text_changed event, always fired on UI thread.
    """

    def __init__(self, document_controller: DocumentController.DocumentController):
        self.__weak_document_controller = typing.cast(_DocumentControllerWeakRefType, weakref.ref(document_controller))
        self.__display_item: typing.Optional[DisplayItem.DisplayItem] = None
        self.__computation: typing.Optional[Symbolic.Computation] = None
        self.__set_display_item(None)
        self.__computation_changed_or_mutated_event_listener: typing.Optional[Event.EventListener] = None
        self.__computation_variable_inserted_event_listener: typing.Optional[Event.EventListener] = None
        self.__computation_variable_removed_event_listener: typing.Optional[Event.EventListener] = None
        self.__computation_property_changed_event_listener: typing.Optional[Event.EventListener] = None
        self.__computation_label: typing.Optional[str] = None
        self.__computation_text: typing.Optional[str] = None
        self.__error_text: typing.Optional[str] = None
        self.__variable_property_changed_event_listeners: typing.Dict[uuid.UUID, Event.EventListener] = dict()
        self.computation_label_changed_event = Event.Event()
        self.computation_text_changed_event = Event.Event()
        self.error_text_changed_event = Event.Event()
        self.variable_inserted_event = Event.Event()
        self.variable_removed_event = Event.Event()
        self.variable_property_changed_event = Event.Event()

    def close(self) -> None:
        self.__set_display_item(None)

    @property
    def document_controller(self) -> DocumentController.DocumentController:
        return self.__weak_document_controller()

    @property
    def display_item(self) -> typing.Optional[DisplayItem.DisplayItem]:
        return self.__display_item

    @property
    def computation(self) -> typing.Optional[Symbolic.Computation]:
        return self.__computation

    def set_display_item(self, display_item: typing.Optional[DisplayItem.DisplayItem]) -> None:
        self.__set_display_item(display_item)

    def add_variable(self, name: typing.Optional[str] = None, value_type: typing.Optional[Symbolic.ComputationVariableType] = None,
                     value: typing.Any = None, value_default: typing.Any = None, value_min: typing.Any = None,
                     value_max: typing.Any = None, control_type: typing.Optional[str] = None,
                     specified_item: typing.Optional[Persistence.PersistentObject] = None,
                     label: typing.Optional[str] = None) -> None:
        computation = self.computation
        if computation:
            command = AddVariableCommand(self.document_controller.document_model, computation, name, value_type, value, value_default, value_min, value_max, control_type, specified_item, label)
            command.perform()
            self.document_controller.push_undo_command(command)

    def remove_variable(self, variable: Symbolic.ComputationVariable) -> None:
        computation = self.computation
        if computation:
            command = RemoveVariableCommand(self.document_controller.document_model, computation, variable)
            command.perform()
            self.document_controller.push_undo_command(command)

    @property
    def computation_label(self) -> typing.Optional[str]:
        return self.__computation_label

    @computation_label.setter
    def computation_label(self, label: typing.Optional[str]) -> None:
        computation = self.computation
        if computation:
            command = ChangeComputationCommand(self.document_controller.document_model, computation, command_id="computation_change_label", is_mergeable=True, label=label)
            command.perform()
            self.document_controller.push_undo_command(command)

    @property
    def computation_text(self) -> typing.Optional[str]:
        return self.__computation_text

    @computation_text.setter
    def computation_text(self, computation_text: typing.Optional[str]) -> None:
        computation = self.computation
        if computation:
            command = ChangeComputationCommand(self.document_controller.document_model, computation, command_id="computation_change_label", is_mergeable=True, expression=computation_text)
            command.perform()
            self.document_controller.push_undo_command(command)

    @property
    def error_text(self) -> typing.Optional[str]:
        return self.__error_text

    def __update_computation_label(self, computation_label: typing.Optional[str]) -> None:
        if self.__computation_label != computation_label:
            self.__computation_label = computation_label
            self.computation_label_changed_event.fire(self.__computation_label)

    def __update_computation_text(self, computation_text: typing.Optional[str]) -> None:
        if self.__computation_text != computation_text:
            self.__computation_text = computation_text
            self.computation_text_changed_event.fire(self.__computation_text)

    def __update_error_text(self, error_text: typing.Optional[str]) -> None:
        if self.__error_text != error_text:
            self.__error_text = error_text
            self.error_text_changed_event.fire(self.__error_text)

    def __update_computation_display(self) -> None:
        def update_computation_display() -> None:
            label = None
            expression = None
            error_text = None
            computation = self.computation
            if computation:
                error_text = computation.error_text
                expression = computation.expression
                label = computation.label
            self.__update_computation_label(label)
            self.__update_computation_text(expression)
            self.__update_error_text(error_text)
        self.document_controller.queue_task(update_computation_display)

    def __computation_item_inserted(self, name: str, index: int, variable: Symbolic.ComputationVariable) -> None:
        if name == "variables":
            self.variable_inserted_event.fire(index, variable)
            def handle_property_changed(key: str) -> None:
                self.__update_computation_display()
                self.variable_property_changed_event.fire(variable)
            self.__variable_property_changed_event_listeners[variable.uuid] = variable.property_changed_event.listen(handle_property_changed)

    def __computation_item_removed(self, name: str, index: int, variable: Symbolic.ComputationVariable) -> None:
        if name == "variables":
            self.variable_removed_event.fire(index, variable)
            self.__variable_property_changed_event_listeners[variable.uuid].close()
            del self.__variable_property_changed_event_listeners[variable.uuid]

    # not thread safe
    def __set_display_item(self, display_item: typing.Optional[DisplayItem.DisplayItem]) -> None:
        if bool(self.__display_item != display_item):
            if self.__computation_changed_or_mutated_event_listener:
                self.__computation_changed_or_mutated_event_listener.close()
                self.__computation_changed_or_mutated_event_listener = None
            if self.__computation_variable_inserted_event_listener:
                self.__computation_variable_inserted_event_listener.close()
                self.__computation_variable_inserted_event_listener = None
            if self.__computation_variable_removed_event_listener:
                self.__computation_variable_removed_event_listener.close()
                self.__computation_variable_removed_event_listener = None
            if self.__computation_property_changed_event_listener:
                self.__computation_property_changed_event_listener.close()
                self.__computation_property_changed_event_listener = None
            computation = self.computation
            if computation:
                for index, variable in enumerate(computation.variables):
                    self.__computation_item_removed("variables", 0, variable)
            document_model = self.document_controller.document_model
            self.__display_item = display_item
            self.__computation = select_computation(document_model, display_item)
            computation = self.computation
            if computation:
                def computation_updated(computation: Symbolic.Computation) -> None:
                    if computation == self.computation:
                        self.__update_computation_display()
                def property_changed(property: str) -> None:
                    if property == "error_text":
                        self.__update_computation_display()
                self.__computation_changed_or_mutated_event_listener = document_model.computation_updated_event.listen(computation_updated)
                self.__computation_variable_inserted_event_listener = computation.item_inserted_event.listen(self.__computation_item_inserted)
                self.__computation_variable_removed_event_listener = computation.item_removed_event.listen(self.__computation_item_removed)
                self.__computation_property_changed_event_listener = computation.property_changed_event.listen(property_changed)
            self.__update_computation_display()
            if computation:
                for index, variable in enumerate(computation.variables):
                    self.__computation_item_inserted("variables", index, variable)


class ChangeVariableBinding(Binding.PropertyBinding):
    def __init__(self, document_controller: DocumentController.DocumentController, computation: Symbolic.Computation,
                 variable: Symbolic.ComputationVariable, property_name: str,
                 converter: typing.Optional[Converter.ConverterLike[typing.Any, typing.Any]] = None,
                 fallback: typing.Any = None) -> None:
        super().__init__(variable, property_name, converter=converter, fallback=fallback)
        self.__document_controller = document_controller
        self.__computation = computation
        self.__variable = variable
        self.__property_name = property_name
        self.__old_source_setter = self.source_setter
        self.source_setter = ReferenceCounting.weak_partial(ChangeVariableBinding.__set_value, self)

    def __set_value(self, value: typing.Any) -> None:
        if value != getattr(self.__variable, self.__property_name):
            command = Inspector.ChangeComputationVariableCommand(self.__document_controller.document_model, self.__computation, typing.cast(Symbolic.ComputationVariable, self.source), **{self.__property_name: value})
            command.perform()
            self.__document_controller.push_undo_command(command)


class ComputationPanelSection:

    def __init__(self, document_controller: DocumentController.DocumentController, computation: Symbolic.Computation, variable: Symbolic.ComputationVariable, on_remove: typing.Optional[typing.Callable[[], None]], queue_task_fn: typing.Callable[[typing.Callable[[], None]], None]) -> None:
        ui = document_controller.ui

        self.variable = variable

        section_widget = ui.create_column_widget()
        section_title_row = ui.create_row_widget()

        twist_down_canvas_item = CanvasItem.TwistDownCanvasItem()
        twist_down_canvas_widget = ui.create_canvas_widget(properties={"height": 20, "width": 20})
        twist_down_canvas_widget.canvas_item.add_canvas_item(twist_down_canvas_item)
        section_title_row.add(twist_down_canvas_widget)
        twist_down_label_widget = ui.create_label_widget(variable.name)
        twist_down_label_widget.text_font = "bold"
        twist_down_label_widget.bind_text(Binding.PropertyBinding(variable, "name"))
        section_title_row.add(twist_down_label_widget)
        section_title_row.add_stretch()
        section_widget.add(section_title_row)

        # boolean, integral, real, data item, region

        def make_label_row(ui: UserInterface.UserInterface, label: str) -> UserInterface.BoxWidget:
            label_row = ui.create_row_widget()
            label_row.add_spacing(8)
            label_row.add(ui.create_label_widget(label))
            label_row.add_stretch()
            return label_row

        def make_name_type_row(ui: UserInterface.UserInterface, variable: Symbolic.ComputationVariable, on_change_type_fn: typing.Callable[[typing.Any], None], on_remove_fn: typing.Optional[typing.Callable[[], None]]) -> UserInterface.BoxWidget:
            name_text_edit = ui.create_line_edit_widget()
            name_text_edit.bind_text(ChangeVariableBinding(document_controller, computation, variable, "name"))

            type_items = [
                (Symbolic.ComputationVariableType.BOOLEAN, _("Boolean")),
                (Symbolic.ComputationVariableType.INTEGRAL, _("Integer")),
                (Symbolic.ComputationVariableType.REAL, _("Real")),
                (Symbolic.ComputationVariableType.STRING, _("String")),
                (Symbolic.ComputationVariableType.DATA_SOURCE, _("Data Source")),
                (Symbolic.ComputationVariableType.GRAPHIC, _("Graphic"))
            ]

            type_combo_box = ui.create_combo_box_widget(items=type_items, item_getter=operator.itemgetter(1))

            remove_button = ui.create_push_button_widget(_("X"))

            name_type_row = ui.create_row_widget()
            name_type_row.add_spacing(8)
            name_type_row.add(name_text_edit)
            name_type_row.add_spacing(8)
            name_type_row.add(type_combo_box)
            name_type_row.add_spacing(8)
            name_type_row.add(remove_button)
            name_type_row.add_stretch()

            variable_type = variable.variable_type
            type_combo_box.current_item = next((i for i in type_items if i[0] == variable_type), None)
            type_combo_box.on_current_item_changed = lambda item: on_change_type_fn(item[0])
            remove_button.on_clicked = on_remove_fn

            return name_type_row

        def make_boolean_row(ui: UserInterface.UserInterface, variable: Symbolic.ComputationVariable, on_change_type_fn: typing.Callable[[typing.Any], None], on_remove_fn: typing.Optional[typing.Callable[[], None]]) -> UserInterface.BoxWidget:
            name_type_row = make_name_type_row(ui, variable, on_change_type_fn, on_remove_fn)

            value_check_box = ui.create_check_box_widget(_("Value"))

            value_row = ui.create_row_widget()
            value_row.add_spacing(8)
            value_row.add(value_check_box)
            value_row.add_stretch()

            label_text_edit = ui.create_line_edit_widget()
            label_text_edit.bind_text(ChangeVariableBinding(document_controller, computation, variable, "label"))

            display_row = ui.create_row_widget()
            display_row.add_spacing(8)
            display_row.add(ui.create_label_widget(_("Label")))
            display_row.add_spacing(8)
            display_row.add(label_text_edit)
            display_row.add_stretch()

            column = ui.create_column_widget()
            column.add(make_label_row(ui, _("Variable Name / Type")))
            column.add(name_type_row)
            column.add(value_row)
            column.add(display_row)

            value_check_box.bind_checked(ChangeVariableBinding(document_controller, computation, variable, "value"))

            return column

        def make_number_row(ui: UserInterface.UserInterface, variable: Symbolic.ComputationVariable, converter: typing.Optional[Converter.ConverterLike[typing.Any, typing.Any]], on_change_type_fn: typing.Callable[[typing.Any], None], on_remove_fn: typing.Optional[typing.Callable[[], None]]) -> UserInterface.BoxWidget:
            name_type_row = make_name_type_row(ui, variable, on_change_type_fn, on_remove_fn)

            value_text_edit = ui.create_line_edit_widget()

            value_default_text_edit = ui.create_line_edit_widget()

            value_min_text_edit = ui.create_line_edit_widget()

            value_max_text_edit = ui.create_line_edit_widget()

            value_row = ui.create_row_widget()
            value_row.add_spacing(8)
            value_row.add(value_text_edit)
            value_row.add_spacing(4)
            value_row.add(value_default_text_edit)
            value_row.add_spacing(4)
            value_row.add(value_min_text_edit)
            value_row.add_spacing(4)
            value_row.add(value_max_text_edit)
            value_row.add_stretch()

            label_text_edit = ui.create_line_edit_widget()
            label_text_edit.bind_text(ChangeVariableBinding(document_controller, computation, variable, "label"))

            display_items = [("field", _("Field")), ("slider", _("Slider"))]
            display_combo_box = ui.create_combo_box_widget(items=display_items, item_getter=operator.itemgetter(1))

            display_row = ui.create_row_widget()
            display_row.add_spacing(8)
            display_row.add(label_text_edit)
            # display_row.add_spacing(4)
            # display_row.add(display_combo_box)
            display_row.add_stretch()

            column = ui.create_column_widget()
            column.add(make_label_row(ui, _("Variable Name / Type")))
            column.add(name_type_row)
            column.add(make_label_row(ui, _("Value / Default / Min / Max")))
            column.add(value_row)
            column.add(make_label_row(ui, _("Label / Display Type")))
            column.add(display_row)

            value_text_edit.bind_text(ChangeVariableBinding(document_controller, computation, variable, "value", converter=converter))
            value_default_text_edit.bind_text(ChangeVariableBinding(document_controller, computation, variable, "value_default", converter=converter))
            value_min_text_edit.bind_text(ChangeVariableBinding(document_controller, computation, variable, "value_min", converter=converter))
            value_max_text_edit.bind_text(ChangeVariableBinding(document_controller, computation, variable, "value_max", converter=converter))

            # display_combo_box.on_current_item_changed ... handle undo
            # display_combo_box.current_item = display_items[1]

            return column

        def make_string_row(ui: UserInterface.UserInterface, variable: Symbolic.ComputationVariable, converter: typing.Optional[Converter.ConverterLike[typing.Any, typing.Any]], on_change_type_fn: typing.Callable[[typing.Any], None], on_remove_fn: typing.Optional[typing.Callable[[], None]]) -> UserInterface.BoxWidget:
            name_type_row = make_name_type_row(ui, variable, on_change_type_fn, on_remove_fn)

            value_text_edit = ui.create_line_edit_widget()

            value_default_text_edit = ui.create_line_edit_widget()

            value_row = ui.create_row_widget()
            value_row.add_spacing(8)
            value_row.add(value_text_edit)
            value_row.add_spacing(4)
            value_row.add(value_default_text_edit)
            value_row.add_stretch()

            label_text_edit = ui.create_line_edit_widget()
            label_text_edit.bind_text(ChangeVariableBinding(document_controller, computation, variable, "label"))

            display_row = ui.create_row_widget()
            display_row.add_spacing(8)
            display_row.add(label_text_edit)
            display_row.add_stretch()

            column = ui.create_column_widget()
            column.add(make_label_row(ui, _("Variable Name / Type")))
            column.add(name_type_row)
            column.add(make_label_row(ui, _("Value / Default")))
            column.add(value_row)
            column.add(make_label_row(ui, _("Label")))
            column.add(display_row)

            value_text_edit.bind_text(ChangeVariableBinding(document_controller, computation, variable, "value", converter=converter))
            value_default_text_edit.bind_text(ChangeVariableBinding(document_controller, computation, variable, "value_default", converter=converter))

            return column

        def make_specifier_row(ui: UserInterface.UserInterface, variable: Symbolic.ComputationVariable, on_change_type_fn: typing.Callable[[typing.Any], None], on_remove_fn: typing.Optional[typing.Callable[[], None]]) -> UserInterface.BoxWidget:
            column = ui.create_column_widget()

            name_type_row = make_name_type_row(ui, variable, on_change_type_fn, on_remove_fn)

            label_text_edit = ui.create_line_edit_widget()
            label_text_edit.bind_text(ChangeVariableBinding(document_controller, computation, variable, "label"))

            display_row = ui.create_row_widget()
            display_row.add_spacing(8)
            display_row.add(ui.create_label_widget(_("Label")))
            display_row.add_spacing(8)
            display_row.add(label_text_edit)
            display_row.add_stretch()

            column.add(name_type_row)
            column.add(display_row)

            return column

        def make_empty_row(ui: UserInterface.UserInterface, variable: Symbolic.ComputationVariable, on_change_type_fn: typing.Callable[[typing.Any], None], on_remove_fn: typing.Optional[typing.Callable[[], None]]) -> UserInterface.BoxWidget:
            column = ui.create_column_widget()
            name_type_row = make_name_type_row(ui, variable, on_change_type_fn, on_remove_fn)
            column.add(name_type_row)
            return column

        stack = ui.create_column_widget()

        section_widget.add(stack)
        section_widget.add_spacing(4)

        def toggle() -> None:
            twist_down_canvas_item.checked = not twist_down_canvas_item.checked
            stack.visible = twist_down_canvas_item.checked
        section_open = False
        twist_down_canvas_item.checked = section_open
        stack.visible = section_open
        twist_down_canvas_item.on_button_clicked = toggle

        def change_type(variable_type: str) -> None:
            command = Inspector.ChangeComputationVariableCommand(document_controller.document_model, computation, variable, title=_("Remove Input Data Item"), variable_type=variable_type)
            command.perform()
            document_controller.push_undo_command(command)

        def select_stack(stack: UserInterface.BoxWidget, variable: Symbolic.ComputationVariable) -> None:
            stack.remove_all()
            variable_type = variable.variable_type
            if variable_type == Symbolic.ComputationVariableType.BOOLEAN:
                stack.add(make_boolean_row(ui, variable, change_type, on_remove))
            elif variable_type == Symbolic.ComputationVariableType.INTEGRAL:
                stack.add(make_number_row(ui, variable, Converter.IntegerToStringConverter(), change_type, on_remove))
            elif variable_type == Symbolic.ComputationVariableType.REAL:
                stack.add(make_number_row(ui, variable, Converter.FloatToStringConverter(), change_type, on_remove))
            elif variable_type == Symbolic.ComputationVariableType.STRING:
                stack.add(make_string_row(ui, variable, None, change_type, on_remove))
            elif variable_type == Symbolic.ComputationVariableType.DATA_SOURCE:
                stack.add(make_specifier_row(ui, variable, change_type, on_remove))
            elif variable_type == Symbolic.ComputationVariableType.GRAPHIC:
                stack.add(make_specifier_row(ui, variable, change_type, on_remove))
            else:
                stack.add(make_empty_row(ui, variable, change_type, on_remove))

        def do_select_stack() -> None:
            # select stack will remove the inspector widgets, so delay it until the
            # current event (combo box changed) has finished by queueing it.
            queue_task_fn(functools.partial(select_stack, stack, variable))

        self.__variable_type_changed_event_listener = variable.variable_type_changed_event.listen(do_select_stack)

        select_stack(stack, variable)

        self.widget = section_widget

    def close(self) -> None:
        self.__variable_type_changed_event_listener.close()
        self.__variable_type_changed_event_listener = typing.cast(typing.Any, None)


def drop_mime_data(document_controller: DocumentController.DocumentController, computation: Symbolic.Computation,
                   variable: Symbolic.ComputationVariable, mime_data: UserInterface.MimeData, x: int, y: int) -> str:
    display_item, graphic = MimeTypes.mime_data_get_data_source(mime_data, document_controller.document_model)
    data_item = display_item.data_item if display_item else None
    if data_item and display_item:
        properties = {"variable_type": "data_source", "secondary_specified_object": graphic, "specified_object": display_item.get_display_data_channel_for_data_item(data_item)}
        command = Inspector.ChangeComputationVariableCommand(document_controller.document_model, computation, variable, title=_("Set Input Data Source"), **properties)  # type: ignore
        command.perform()
        document_controller.push_undo_command(command)
        return "copy"
    display_item = MimeTypes.mime_data_get_display_item(mime_data, document_controller.document_model)
    data_item = display_item.data_item if display_item else None
    if data_item and display_item:
        properties = {"variable_type": "data_source", "secondary_specified_object": None, "specified_object": display_item.get_display_data_channel_for_data_item(data_item)}
        command = Inspector.ChangeComputationVariableCommand(document_controller.document_model, computation, variable, title=_("Set Input Data Source"), **properties)  # type: ignore
        command.perform()
        document_controller.push_undo_command(command)
        return "copy"
    return "ignore"


def data_item_delete(document_controller: DocumentController.DocumentController, computation: Symbolic.Computation, variable: Symbolic.ComputationVariable) -> None:
    command = Inspector.ChangeComputationVariableCommand(document_controller.document_model, computation, variable, title=_("Remove Input Data Source"), specified_object=None)
    command.perform()
    document_controller.push_undo_command(command)


def make_image_chooser(document_controller: DocumentController.DocumentController,
                       computation: Symbolic.Computation,
                       variable: Symbolic.ComputationVariable,
                       drag_fn: typing.Callable[[UserInterface.MimeData, typing.Optional[DrawingContext.RGBA32Type], typing.Optional[int], typing.Optional[int]], None]) -> typing.Tuple[UserInterface.BoxWidget, typing.Sequence[Event.EventListener]]:
    ui = document_controller.ui
    document_model = document_controller.document_model
    # drag_fn is necessary because it is unsafe to start a drag on the column containing the thumbnail
    # since dragging onto itself may delete the column during the drag.
    column = ui.create_column_widget(properties={"width": 80})
    label_row = ui.create_row_widget()
    label_widget = ui.create_label_widget(variable.display_label, properties={"width": 80})
    label_widget.bind_text(Binding.PropertyBinding(variable, "display_label"))
    label_row.add_stretch()
    label_row.add(label_widget)
    label_row.add_stretch()
    data_item = computation.get_input_data_item(variable.name)

    display_item = document_model.get_display_item_for_data_item(data_item) if data_item else None
    data_item_thumbnail_source = DataItemThumbnailWidget.DataItemThumbnailSource(ui, display_item=display_item)
    data_item_chooser_widget = DataItemThumbnailWidget.ThumbnailWidget(ui, data_item_thumbnail_source, Geometry.IntSize(80, 80))

    def thumbnail_widget_drag(mime_data: UserInterface.MimeData, thumbnail: typing.Optional[DrawingContext.RGBA32Type], hot_spot_x: int, hot_spot_y: int) -> None:
        # use this convoluted base object for drag so that it doesn't disappear after the drag.
        drag_fn(mime_data, thumbnail, hot_spot_x, hot_spot_y)

    data_item_chooser_widget.on_drag = thumbnail_widget_drag
    data_item_chooser_widget.on_delete = functools.partial(data_item_delete, document_controller, computation, variable)
    data_item_chooser_widget.on_drop_mime_data = functools.partial(drop_mime_data, document_controller, computation, variable)

    def property_changed(key: str) -> None:
        if key == "specified_object":
            data_item = computation.get_input_data_item(variable.name)
            display_item = document_model.get_display_item_for_data_item(data_item) if data_item else None
            if display_item:
                data_item_thumbnail_source.set_display_item(display_item)

    property_changed_listener = variable.property_changed_event.listen(property_changed)
    column.add_spacing(4)
    column.add(data_item_chooser_widget)
    column.add(label_row)
    column.add_spacing(4)
    return column, [property_changed_listener]


class EditComputationDialog(Dialog.ActionDialog):

    def __init__(self, document_controller: DocumentController.DocumentController, data_item: DataItem.DataItem):

        display_item = document_controller.document_model.get_display_item_for_data_item(data_item)
        assert display_item

        ui = document_controller.ui
        super().__init__(ui, _("Edit Computation"), parent_window=document_controller, persistent_id="EditComputationDialog" + str(display_item.uuid))

        self.ui = ui
        self.document_controller = document_controller

        self.__computation_model = ComputationModel(document_controller)

        self.__sections: typing.List[ComputationPanelSection] = list()

        label_edit_widget = ui.create_line_edit_widget(properties={"min-width": 120})
        label_edit_widget.placeholder_text = _("Computation Label")

        label_row = ui.create_row_widget()
        label_row.add_spacing(8)
        label_row.add(ui.create_label_widget(_("Description")))
        label_row.add_spacing(8)
        label_row.add(label_edit_widget)
        label_row.add_spacing(8)
        label_row.add_stretch()

        buttons_row = ui.create_row_widget()
        add_variable_button = ui.create_push_button_widget(_("Add Variable"))
        add_object_button = ui.create_push_button_widget(_("Add Object"))
        buttons_row.add(add_variable_button)
        buttons_row.add_spacing(8)
        buttons_row.add(add_object_button)
        buttons_row.add_spacing(8)
        buttons_row.add_stretch()

        # sizing in widget space (Qt) is difficult to get right and there seems to be bugs.
        # in this case, two different elements are used to effectively make a minimum window
        # size -- the text edit widget for the height and the error row for the width.
        # if both are done on the text edit widget itself, which would be preferred, Qt seems
        # to give up on layout when the scroll bar appears for too many lines.

        text_edit_row = ui.create_row_widget()
        text_edit = ui.create_text_edit_widget(properties={"min-height": 180})
        text_edit.placeholder = _("No Computation")
        text_edit_row.add_spacing(8)
        text_edit_row.add(text_edit)
        text_edit_row.add_spacing(8)

        error_row = ui.create_row_widget(properties={"min-width": 400})  # the stylesheet allows it to shrink. guh.
        error_label = ui.create_label_widget("\n", properties={"min-width": 120})
        error_label.text_color = "red"
        error_label.word_wrap = True
        error_row.add_spacing(8)
        error_row.add(error_label)
        error_row.add_spacing(8)

        self.__text_edit = text_edit  # for testing
        self.__error_label = error_label  # for testing

        update_button = ui.create_push_button_widget(_("Update"))
        button_row = ui.create_row_widget()
        button_row.add_stretch()
        button_row.add(update_button)
        button_row.add_spacing(8)

        self.__data_item_row = ui.create_row_widget()

        self.__variable_column = ui.create_column_widget()

        def add_object_pressed() -> None:
            document_model = document_controller.document_model
            self.__computation_model.add_variable("".join([random.choice(string.ascii_lowercase) for _ in range(4)]), specified_item=document_model.data_items[0])

        add_object_button.on_clicked = add_object_pressed

        def add_variable_pressed() -> None:
            self.__computation_model.add_variable("".join([random.choice(string.ascii_lowercase) for _ in range(4)]), value_type=Symbolic.ComputationVariableType.INTEGRAL, value=0)

        add_variable_button.on_clicked = add_variable_pressed

        def update_pressed() -> None:
            if text_edit.text:
                self.__computation_model.computation_text = text_edit.text

        update_button.on_clicked = update_pressed
        def editing_finished(text: str) -> None:
            if self.__computation_model:
                self.__computation_model.computation_label = text
        label_edit_widget.on_editing_finished = editing_finished

        def computation_label_changed(text: str) -> None:
            label_edit_widget.text = text
            if label_edit_widget.focused:
                label_edit_widget.request_refocus()

        self.__computation_label_changed_event_listener = self.__computation_model.computation_label_changed_event.listen(computation_label_changed)

        def computation_text_changed(computation_text: str) -> None:
            text_edit.text = computation_text

        self.__computation_text_changed_event_listener = self.__computation_model.computation_text_changed_event.listen(computation_text_changed)

        def error_text_changed(error_text: str) -> None:
            error_label.text = error_text

        self.__error_text_changed_event_listener = self.__computation_model.error_text_changed_event.listen(error_text_changed)

        self.__listeners: typing.List[Event.EventListener] = list()

        def rebuild_data_item_row() -> None:
            self.__data_item_row.remove_all()
            for listener in self.__listeners:
                listener.close()
            self.__listeners = list()
            self.__data_item_row.add_spacing(8)
            for section in self.__sections:
                variable = section.variable
                if variable.variable_type == Symbolic.ComputationVariableType.DATA_SOURCE:
                    computation = self.__computation_model.computation
                    assert computation
                    widget, listeners = make_image_chooser(document_controller, computation, variable, self.content.drag)
                    self.__listeners.extend(listeners)
                    self.__data_item_row.add(widget)
                    self.__data_item_row.add_spacing(8)
            self.__data_item_row.add_stretch()
            target_column = ui.create_column_widget(properties={"width": 80})

            def thumbnail_widget_drag(mime_data: UserInterface.MimeData, thumbnail: typing.Optional[DrawingContext.RGBA32Type], hot_spot_x: int, hot_spot_y: int) -> None:
                # use this convoluted base object for drag so that it doesn't disappear after the drag.
                self.content.drag(mime_data, thumbnail, hot_spot_x, hot_spot_y)

            data_item_thumbnail_source = DataItemThumbnailWidget.DataItemThumbnailSource(ui, display_item=display_item)  # TODO: never closed
            data_item_chooser_widget = DataItemThumbnailWidget.ThumbnailWidget(ui, data_item_thumbnail_source, Geometry.IntSize(80, 80))
            data_item_chooser_widget.on_drag = thumbnail_widget_drag
            target_column.add_spacing(4)
            target_column.add(data_item_chooser_widget)
            target_column.add(ui.create_label_widget(_("Target"), properties={"width": 80}))
            target_column.add_spacing(4)
            self.__data_item_row.add(target_column)
            self.__data_item_row.add_spacing(8)

        def variable_inserted(index: int, variable: Symbolic.ComputationVariable) -> None:
            def remove_variable() -> None:
                self.__computation_model.remove_variable(variable)

            computation = self.__computation_model.computation
            assert computation
            section = ComputationPanelSection(document_controller, computation, variable, remove_variable, self.document_controller.queue_task)
            self.__variable_column.insert(section.widget, index)
            self.__sections.insert(index, section)
            rebuild_data_item_row()

        def variable_removed(index: int, variable: Symbolic.ComputationVariable) -> None:
            self.__variable_column.remove(self.__variable_column.children[index])
            if self.__sections[index]:
                self.__sections[index].close()
            del self.__sections[index]
            rebuild_data_item_row()

        def variable_property_changed(variable: Symbolic.ComputationVariable) -> None:
            rebuild_data_item_row()

        self.__variable_inserted_event_listener = self.__computation_model.variable_inserted_event.listen(variable_inserted)
        self.__variable_removed_event_listener = self.__computation_model.variable_removed_event.listen(variable_removed)
        self.__variable_property_changed_event_listener = self.__computation_model.variable_property_changed_event.listen(variable_property_changed)

        # for testing
        self._update_button = update_button

        self.__computation_model.set_display_item(display_item)

        column = self.content
        column.add_spacing(6)
        column.add(label_row)
        column.add(self.__data_item_row)
        column.add(self.__variable_column)
        column.add(buttons_row)
        column.add_spacing(6)
        column.add(text_edit_row)
        column.add_spacing(6)
        column.add(error_row)
        column.add_spacing(6)
        column.add(button_row)
        column.add_spacing(6)

    def close(self) -> None:
        self.document_controller.clear_task(str(id(self)))
        for listener in self.__listeners:
            listener.close()
        self.__listeners = list()
        self.__computation_label_changed_event_listener.close()
        self.__computation_label_changed_event_listener = typing.cast(typing.Any, None)
        self.__computation_text_changed_event_listener.close()
        self.__computation_text_changed_event_listener = typing.cast(typing.Any, None)
        self.__error_text_changed_event_listener.close()
        self.__error_text_changed_event_listener = typing.cast(typing.Any, None)
        self.__variable_inserted_event_listener.close()
        self.__variable_inserted_event_listener = typing.cast(typing.Any, None)
        self.__variable_removed_event_listener.close()
        self.__variable_removed_event_listener = typing.cast(typing.Any, None)
        self.__computation_model.close()
        self.__computation_model = typing.cast(typing.Any, None)
        super().close()

    def size_changed(self, width: int, height: int) -> None:
        self.__error_label.size = Geometry.IntSize(height=self.__error_label.size.height, width=self.__text_edit.size.width)

    @property
    def _sections_for_testing(self) -> typing.Sequence[ComputationPanelSection]:
        return self.__sections

    @property
    def _variable_column_for_testing(self) -> UserInterface.BoxWidget:
        return self.__variable_column

    @property
    def _text_edit_for_testing(self) -> UserInterface.TextEditWidget:
        return self.__text_edit

    @property
    def _error_label_for_testing(self) -> UserInterface.LabelWidget:
        return self.__error_label

    @property
    def _computation_model_for_testing(self) -> ComputationModel:
        return self.__computation_model


class VariableHandler(Declarative.Handler):
    def __init__(self, computation_inspector_context: Inspector.ComputationInspectorContext, computation: Symbolic.Computation, variable: Symbolic.ComputationVariable):
        super().__init__()
        self.computation = computation
        self.variable = variable
        variable_model = Inspector.VariableValueModel(computation_inspector_context.window, computation, variable)
        # use 2000 below to avoid a match slider, which gives rise to a bizarre slider bug https://bugreports.qt.io/browse/QTBUG-77368
        # also must be a multiple of inspector slider to avoid
        self.slider_converter = Converter.FloatToScaledIntegerConverter(2000, 0, 100)
        self.float_str_converter = Converter.FloatToStringConverter()
        self.int_str_converter = Converter.IntegerToStringConverter()
        self.property_changed_event = Event.Event()
        self.__variable_component: typing.Optional[Declarative.HandlerLike] = Inspector.make_computation_variable_component(computation_inspector_context, computation, variable, variable_model)
        u = Declarative.DeclarativeUI()
        if self.__variable_component:
            self.ui_view = u.create_column(u.create_component_instance("component"), spacing=8)
        else:
            label = u.create_label(text="@binding(variable.display_label)")
            self.ui_view = u.create_column(label, u.create_label(text=_("Missing") + " " + f"[{variable.variable_type}]"), spacing=8)

    def create_handler(self, component_id: str, container: typing.Optional[Symbolic.ComputationVariable] = None, item: typing.Any = None, **kwargs: typing.Any) -> typing.Optional[Declarative.HandlerLike]:
        if component_id == "component":
            assert self.__variable_component
            return self.__variable_component
        return None


class ResultHandler(Declarative.Handler):
    def __init__(self, document_controller: DocumentController.DocumentController, computation: Symbolic.Computation, result: Symbolic.ComputationOutput):
        super().__init__()
        self.document_controller = document_controller
        self.computation = computation
        self.result = result
        self.ui_view = self.__make_component_content(result)

    @property
    def display_item(self) -> typing.Optional[DisplayItem.DisplayItem]:
        document_model = self.document_controller.document_model
        result_items = self.result.output_items
        if len(result_items) == 1 and isinstance(result_items[0], DataItem.DataItem):
            return document_model.get_best_display_item_for_data_item(result_items[0])
        return None

    def __make_component_content(self, result: Symbolic.ComputationOutput) -> Declarative.UIDescription:
        u = Declarative.DeclarativeUI()
        label = u.create_label(text="@binding(result.label)")
        result_items = result.output_items
        if len(result_items) == 1 and isinstance(result_items[0], DataItem.DataItem):
            data_source_chooser = {
                "type": "data_source_chooser",
                "display_item": "@binding(display_item)",
                "min_width": 80,
                "min_height": 80,
            }
            return u.create_column(label, data_source_chooser, spacing=8)
        return u.create_column(label)


class RemoveComputationCommand(Undo.UndoableCommand):

    def __init__(self, document_controller: DocumentController.DocumentController, computation: Symbolic.Computation):
        super().__init__(_("Remove Computation"))
        self.__document_controller = document_controller
        workspace_controller = self.__document_controller.workspace_controller
        assert workspace_controller
        self.__old_workspace_layout: typing.Optional[Persistence.PersistentDictType] = workspace_controller.deconstruct()
        self.__new_workspace_layout: typing.Optional[Persistence.PersistentDictType] = None
        self.__computation_index = document_controller.document_model.computations.index(computation)
        self.__undelete_logs: typing.List[Changes.UndeleteLog] = list()
        self.initialize()

    def close(self) -> None:
        self.__document_controller = typing.cast(typing.Any, None)
        self.__old_workspace_layout = None
        self.__new_workspace_layout = None
        for undelete_log in self.__undelete_logs:
            undelete_log.close()
        self.__undelete_logs = typing.cast(typing.Any, None)
        super().close()

    def _perform(self) -> None:
        document_model = self.__document_controller.document_model
        computation = document_model.computations[self.__computation_index]
        self.__undelete_logs.append(document_model.remove_computation_with_log(computation))

    def _get_modified_state(self) -> typing.Any:
        return self.__document_controller.document_model.modified_state

    def _set_modified_state(self, modified_state: typing.Any) -> None:
        self.__document_controller.document_model.modified_state = modified_state

    def _undo(self) -> None:
        workspace_controller = self.__document_controller.workspace_controller
        if workspace_controller:
            self.__new_workspace_layout = workspace_controller.deconstruct()
            for undelete_log in reversed(self.__undelete_logs):
                self.__document_controller.document_model.undelete_all(undelete_log)
                undelete_log.close()
            self.__undelete_logs.clear()
            if self.__old_workspace_layout is not None:
                workspace_controller.reconstruct(self.__old_workspace_layout)

    def _redo(self) -> None:
        self.perform()
        workspace_controller = self.__document_controller.workspace_controller
        if workspace_controller and self.__new_workspace_layout is not None:
            workspace_controller.reconstruct(self.__new_workspace_layout)


class ComputationInspectorModel(Observable.Observable):
    def __init__(self, computation: Symbolic.Computation) -> None:
        super().__init__()
        self.computation = computation
        self.computation_inputs_model = ListModel.FilteredListModel(container=computation, master_items_key="variables")
        self.computation_inputs_model.filter = ListModel.PredicateFilter(lambda v: v.variable_type in Symbolic._data_source_types)
        self.computation_parameters_model = ListModel.FilteredListModel(container=computation, master_items_key="variables")
        self.computation_parameters_model.filter = ListModel.PredicateFilter(lambda v: v.variable_type not in Symbolic._data_source_types)
        self.is_custom = computation.expression is not None
        self.error_state_model = Model.PropertyModel(0)
        self.__error_state_listener = computation.property_changed_event.listen(ReferenceCounting.weak_partial(ComputationInspectorModel.__property_changed, self))
        self.__is_error = self.computation.error_text is not None and self.computation.error_text != str()

    def close(self) -> None:
        self.computation_inputs_model.close()
        self.computation_inputs_model = typing.cast(typing.Any, None)
        self.computation_parameters_model.close()
        self.computation_parameters_model = typing.cast(typing.Any, None)

    def finish_init(self) -> None:
        self.__property_changed("error_text")

    def __property_changed(self, key: str) -> None:
        if key == "error_text":
            self.__is_error = self.computation.error_text is not None and self.computation.error_text != str()
            self.notify_property_changed("status_color")
            self.error_state_model.value = 1 if self.__is_error else 0

    @property
    def status_color(self) -> str:
        return "red" if self.__is_error else "black"


class ComputationInspectorHandler(Declarative.Handler):
    def __init__(self, context: EntityBrowser.Context, computation: Symbolic.Computation):
        super().__init__()
        assert isinstance(context, Inspector.ComputationInspectorContext)
        self.context = context
        self.document_controller = typing.cast("DocumentController.DocumentController", context.values["window"])
        self.model = ComputationInspectorModel(computation)
        self.delete_state_model = Model.PropertyModel(0)
        self.ui_view = self.__make_ui()
        self.model.finish_init()

    def close(self) -> None:
        self.model.close()
        self.model = typing.cast(typing.Any, None)
        super().close()

    def __make_ui(self) -> Declarative.UIDescriptionResult:
        u = Declarative.DeclarativeUI()
        label = u.create_label(text=self.model.computation.label)
        source_line = [u.create_component_instance("source_component")] if self.context.values.get("do_references", False) else []
        status = u.create_label(text="@binding(model.computation.status)", color="@binding(model.status_color)", max_width=300)
        inputs = u.create_column(items="model.computation_inputs_model.items", item_component_id="variable", spacing=8, size_policy_vertical="expanding")
        results = u.create_column(items="model.computation.results", item_component_id="result", spacing=8, size_policy_vertical="expanding")
        input_output_row = u.create_row(
            u.create_column(u.create_column(inputs), u.create_stretch()),
            u.create_column(u.create_column(results), u.create_stretch()),
            spacing=12,
            size_policy_vertical="expanding"
        )
        parameters = u.create_column(items="model.computation_parameters_model.items", item_component_id="variable", spacing=8)
        if sys.platform == "darwin":
            note = u.create_row(u.create_label(text=_("Use Command+Shift+E to edit data item script.")), visible="@binding(model.is_custom)")
        else:
            note = u.create_row(u.create_label(text=_("Use Ctrl+Shift+E to edit data item script.")), visible="@binding(model.is_custom)")
        controls = u.create_row(u.create_column(status, note, u.create_stretch(), spacing=12), u.create_stretch())
        inspector_column = u.create_column(label, *source_line, u.create_column(input_output_row, parameters, u.create_divider(orientation="horizontal"), controls, spacing=12), spacing=12)
        return inspector_column

    def create_handler(self, component_id: str, container: typing.Optional[Symbolic.ComputationVariable] = None, item: typing.Any = None, **kwargs: typing.Any) -> typing.Optional[Declarative.HandlerLike]:
        if component_id == "variable":
            return VariableHandler(self.context, self.model.computation, typing.cast(Symbolic.ComputationVariable, item))
        elif component_id == "result":
            return ResultHandler(self.document_controller, self.model.computation, typing.cast(Symbolic.ComputationOutput, item))
        elif component_id == "source_component":
            return EntityBrowser.ReferenceHandler(self.context, _("Source"), self.model.computation.source)
        return None


class ComputationErrorInspectorHandler(Declarative.Handler):
    def __init__(self, context: EntityBrowser.Context, computation: Symbolic.Computation):
        super().__init__()
        self.context = context
        self.model = ComputationInspectorModel(computation)
        self.ui_view = self.__make_ui()
        self.model.finish_init()

    def close(self) -> None:
        self.model.close()
        self.model = typing.cast(typing.Any, None)
        super().close()

    def __make_ui(self) -> Declarative.UIDescriptionResult:
        u = Declarative.DeclarativeUI()
        error_column = u.create_column(
            u.create_stack(
                u.create_column(u.create_row(u.create_label(text=_("No Error")), u.create_stretch()), u.create_stretch()),
                u.create_column(u.create_row(u.create_label(text=_("Error")), u.create_stretch()),
                                u.create_row(u.create_label(text="@binding(model.computation.status)", color="@binding(model.status_color)", max_width=300), u.create_stretch()),
                                u.create_text_edit(editable=False, placeholder_text=_("No stack trace."), text="@binding(model.computation.error_stack_trace)"),
                                u.create_stretch(),
                                spacing=8),
                current_index="@binding(model.error_state_model.value)"
            )
        )
        return error_column


class ComputationHandler(Declarative.Handler):
    def __init__(self, document_controller: DocumentController.DocumentController, computation: Symbolic.Computation):
        super().__init__()
        self.context = Inspector.ComputationInspectorContext(document_controller, document_controller, False)
        self.computation = computation
        u = Declarative.DeclarativeUI()
        self.ui_view = u.create_tabs(
            u.create_tab(_("Edit"), u.create_component_instance("edit")),
            u.create_tab(_("Errors"), u.create_component_instance("errors")),
            style="minimal"
        )

    def create_handler(self, component_id: str, container: typing.Optional[Symbolic.ComputationVariable] = None, item: typing.Any = None, **kwargs: typing.Any) -> typing.Optional[Declarative.HandlerLike]:
        if component_id == "edit":
            return ComputationInspectorHandler(self.context, self.computation)
        if component_id == "errors":
            return ComputationErrorInspectorHandler(self.context, self.computation)
        return None


class InspectComputationDialog(Declarative.WindowHandler):
    # handle computations being removed - inspector contents may become invalid
    # implement a chooser for which computation. order by graphic selection first.
    # error display
    # use api computation?
    # improve window placement to be next to selected display panel or data panel
    # data sources should show crop/interval display/editor if present
    # data sources should indicate masking if used
    # data sources should tool tip info about name, size, data type, calibrated size, etc.
    # fall back to description registered for processing id (i.e. label, type, etc.)
    # description/help for overall function and individual parameters
    # dynamic combo box
    # progress bar

    def __init__(self, document_controller: DocumentController.DocumentController, computation: Symbolic.Computation) -> None:
        super().__init__()

        self.__document_controller = document_controller

        # close any previous computation inspector associated with the window
        previous_window = getattr(document_controller, "_computation_inspector", None)
        if isinstance(previous_window, InspectComputationDialog):
            previous_window.close_window()
        setattr(document_controller, "_computation_inspector", self)

        # define models that manage the state of the UI
        self.stack_index_model = Model.PropertyModel(0)
        self.stack_page_model = Model.PropertyModel(str())
        self.computation_model = Model.PropertyModel[Symbolic.Computation]()

        # configure the models
        self.computation_model.value = computation
        self.stack_index_model.value = 1 if self.computation_model.value else 0
        self.stack_page_model.value = ["empty", "single"][self.stack_index_model.value]

        # list for computation being removed
        def remove_computation() -> None:
            self.computation_model.value = None
            self.stack_index_model.value = 0
            self.stack_page_model.value = "empty"

        self.__computation_about_to_be_removed_event_listener = computation.about_to_be_removed_event.listen(remove_computation) if computation else None

        u = Declarative.DeclarativeUI()
        main_page = u.create_column(u.create_component_instance("@binding(stack_page_model.value)"), u.create_stretch())
        window = u.create_window(main_page, title=_("Computation"), margin=12, window_style="tool")
        self.run(window, parent_window=document_controller, persistent_id="computation_inspector")

    def close(self) -> None:
        if self.__computation_about_to_be_removed_event_listener:
            self.__computation_about_to_be_removed_event_listener.close()
            self.__computation_about_to_be_removed_event_listener = None
        setattr(self.__document_controller, "_computation_inspector", None)
        super().close()

    def get_resource(self, resource_id: str, container: typing.Optional[Symbolic.ComputationVariable] = None, item: typing.Any = None) -> typing.Optional[Declarative.UIDescription]:
        if resource_id == "empty":
            u = Declarative.DeclarativeUI()
            content = u.create_column(u.create_label(text=_("No computation.")))
            component = u.define_component(content=content)
            return component
        return None

    def create_handler(self, component_id: str, container: typing.Optional[Symbolic.ComputationVariable] = None, item: typing.Any = None, **kwargs: typing.Any) -> typing.Optional[Declarative.HandlerLike]:
        if component_id == "empty":
            return Declarative.Handler()
        if component_id == "single":
            computation = self.computation_model.value
            assert computation
            return ComputationHandler(self.__document_controller, computation)
        return None


class ComputationErrorNotificationHandler(Declarative.Handler):
    """Declarative component handler for a section in a multiple acquire method component."""

    def __init__(self, app: Application.BaseApplication, notification: Symbolic.ComputationErrorNotification) -> None:
        super().__init__()
        self.app = app
        self.notification = notification
        u = Declarative.DeclarativeUI()
        self.ui_view = u.create_row(
            u.create_column(
                u.create_row(
                    u.create_label(text="@binding(notification.task_name)", color="#3366CC"),
                    u.create_stretch(),
                    {"type": "notification_char_button", "text": " \N{MULTIPLICATION X} ", "on_clicked": "handle_dismiss"},
                    spacing=8
                ),
                u.create_row(
                    {"type": "notification_char_button", "text": "Edit Computation...", "on_clicked": "handle_edit"},
                    u.create_stretch(),
                    spacing=8
                ),
                u.create_row(
                    u.create_label(text="@binding(notification.text)", word_wrap=True, width=440),
                    u.create_stretch(),
                    spacing=8
                ),
                u.create_divider(orientation="horizontal"),
                spacing=4,
            ),
        )

    def handle_dismiss(self, widget: Declarative.UIWidget) -> None:
        self.notification.dismiss()

    def handle_edit(self, widget: Declarative.UIWidget) -> None:
        computation = self.notification.computation
        for window in self.app.windows:
            document_model = typing.cast(typing.Optional["DocumentModel.DocumentModel"], getattr(window, "document_model", None))
            if document_model and computation in document_model.computations:
                InspectComputationDialog(typing.cast("DocumentController.DocumentController", window), computation)


class ComputationErrorNotificationComponentFactory(NotificationDialog.NotificationComponentFactory):
    def make_component(self, app: Application.BaseApplication, notification: Notification.Notification) -> typing.Optional[Declarative.HandlerLike]:
        if isinstance(notification, Symbolic.ComputationErrorNotification):
            return ComputationErrorNotificationHandler(app, notification)
        return None

Registry.register_component(ComputationErrorNotificationComponentFactory(), {"notification-component-factory"})



from __future__ import annotations

import abc
import collections
import copy
import datetime
import json
import os
import pathlib
import typing
import uuid
import weakref

from nion.utils import Converter
from nion.utils import DateTime
from nion.utils import Observable

DictValue = typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Any], typing.Tuple[typing.Any], str, float, int, bool, None]
ItemProxyEntity = typing.Optional[typing.Any]  # use this to make it easy to switch to Entity later
PersistentDictType = typing.Dict[str, DictValue]
PersistentMappingType = typing.Mapping[str, DictValue]
_WeakReferenceType = typing.Any  # Python 3.9+ fix these

ANY = "any"
STRING = "string"
BOOLEAN = "boolean"
INT = "integer"
FLOAT = "double"
LIST = "list"
DICT = "dict"
SET = "set"
TIMESTAMP = "timestamp"
UUID = "uuid"
PATH = "path"

OPTIONAL = True
REQUIRED = False

entity_types: typing.Dict[str, EntityType] = dict()


def register_entity_type(entity_id: str, entity: EntityType) -> None:
    assert entity_id not in entity_types
    entity_types[entity_id] = entity


def unregister_entity_type(entity_id: str) -> None:
    del entity_types[entity_id]


def get_entity_type(entity_id: str) -> typing.Optional[EntityType]:
    return entity_types.get(entity_id)


def build_value(type: str, value: typing.Any) -> typing.Any:
    if value is None:
        return None
    if type in (SET,):
        return set(value)
    if type in (BOOLEAN,):
        return bool(value)
    if type in (INT,):
        return int(value)
    if type in (FLOAT,):
        return float(value)
    if type in (TIMESTAMP,):
        return Converter.DatetimeToStringConverter().convert_back(value)
    if type in (UUID,):
        return Converter.UuidToStringConverter().convert_back(value)
    if type in (PATH,):
        return Converter.PathToStringConverter().convert_back(value)
    return value


def dict_value(type: str, value: typing.Any) -> DictValue:
    if value is None:
        return None
    if type in (SET,):
        return list(value)
    if type in (BOOLEAN,):
        return bool(value)
    if type in (INT,):
        return int(value)
    if type in (FLOAT,):
        return float(value)
    if type in (TIMESTAMP,):
        return Converter.DatetimeToStringConverter().convert(value)
    if type in (UUID,):
        return Converter.UuidToStringConverter().convert(value)
    if type in (PATH,):
        return Converter.PathToStringConverter().convert(value)
    return typing.cast(DictValue, value)


class ItemProxy:
    def __init__(self, item_uuid: typing.Optional[uuid.UUID]) -> None:
        self.__item_uuid = item_uuid
        self.__item = None

    def close(self) -> None:
        pass

    @property
    def item_uuid(self) -> typing.Optional[uuid.UUID]:
        return self.__item_uuid

    @item_uuid.setter
    def item_uuid(self, value: typing.Optional[uuid.UUID]) -> None:
        self.__item_uuid = value

    @property
    def item(self) -> ItemProxyEntity:
        return self.__item

    @item.setter
    def item(self, item: ItemProxyEntity) -> None:
        self.__item = item


class Accessor(typing.Protocol):
    accessor: typing.Optional[Accessor]
    field_type: FieldType

    def get_value(self, item: typing.Any) -> typing.Any: ...

    def breadcrumbs(self, item: typing.Any) -> typing.Sequence[typing.Any]: ...


class BaseAccessor(Accessor):
    def __init__(self, field_type: FieldType, field_name: str) -> None:
        self.accessor = None
        self.field_type = field_type
        self.field_name = field_name

    def __repr__(self) -> str:
        return f"{self.field_name}"

    def get_value(self, item: typing.Any) -> typing.Any:
        return item

    def breadcrumbs(self, item: typing.Any) -> typing.Sequence[typing.Any]:
        return [item]


class FieldAccessor(Accessor):
    def __init__(self, accessor: Accessor, field_type: FieldType, field_name: str) -> None:
        self.accessor = accessor
        self.field_type = field_type
        self.field_name = field_name

    def __repr__(self) -> str:
        return f"{self.accessor}.{self.field_name}"

    def get_value(self, item: typing.Any) -> typing.Any:
        return getattr(self.accessor.get_value(item), self.field_name, None)

    def breadcrumbs(self, item: typing.Any) -> typing.Sequence[typing.Any]:
        return self.accessor.breadcrumbs(item)


class IndexAccessor(Accessor):
    def __init__(self, accessor: Accessor, field_type: FieldType, index: int) -> None:
        self.accessor = accessor
        self.field_type = field_type
        self.index = index

    def __repr__(self) -> str:
        return f"{self.accessor}[{self.index}]"

    def get_value(self, item: typing.Any) -> typing.Any:
        return self.accessor.get_value(item)[self.index]

    def breadcrumbs(self, item: typing.Any) -> typing.Sequence[typing.Any]:
        return list(self.accessor.breadcrumbs(item)) + [self.get_value(item)]


class MapAccessor(Accessor):
    def __init__(self, accessor: Accessor, field_type: FieldType, key: typing.Any) -> None:
        self.accessor = accessor
        self.field_type = field_type
        self.key = key

    def __repr__(self) -> str:
        return f"{self.accessor}['{self.key}']"

    def get_value(self, item: typing.Any) -> typing.Any:
        return self.accessor.get_value(item)[self.key]

    def breadcrumbs(self, item: typing.Any) -> typing.Sequence[typing.Any]:
        return list(self.accessor.breadcrumbs(item)) + [self.get_value(item)]


class Visitor(typing.Protocol):
    def visit(self, accessor: Accessor) -> None: ...


class SupportsVisit(typing.Protocol):
    def visit(self, value: typing.Any, accessor: Accessor, visitor: Visitor) -> None: ...


class EntityContext(abc.ABC):

    @abc.abstractmethod
    def register(self, entity: Entity) -> None: ...

    @abc.abstractmethod
    def unregister(self, entity: Entity) -> None: ...

    @abc.abstractmethod
    def create_item_reference(self, item_uuid: typing.Optional[uuid.UUID] = None) -> ItemProxy: ...

    @abc.abstractmethod
    def update_item_reference_uuid(self, item_reference: ItemProxy, item_uuid: typing.Optional[uuid.UUID]) -> None: ...


class SimpleEntityContext(EntityContext):
    def __init__(self) -> None:
        self.__entity_map: typing.Dict[uuid.UUID, _WeakReferenceType] = dict()
        self.__item_proxies: typing.List[_WeakReferenceType] = list()

    def register(self, entity: Entity) -> None:
        self.__entity_map[entity.uuid] = weakref.ref(entity)
        for item_proxy_ref in self.__item_proxies:
            item_proxy = item_proxy_ref()
            if item_proxy and item_proxy.item_uuid == entity.uuid:
                item_proxy.item = entity

    def unregister(self, entity: Entity) -> None:
        self.__entity_map.pop(entity.uuid)
        for item_proxy_ref in self.__item_proxies:
            item_proxy = item_proxy_ref()
            if item_proxy and item_proxy.item_uuid == entity.uuid:
                item_proxy.item = None

    def create_item_reference(self, item_uuid: typing.Optional[uuid.UUID] = None) -> ItemProxy:
        item_proxy = ItemProxy(item_uuid)
        if item_uuid and item_uuid in self.__entity_map:
            entity_ref = self.__entity_map.get(item_uuid)
            if entity_ref:
                item_proxy.item = entity_ref()
        item_proxy_ref = weakref.ref(item_proxy)
        self.__item_proxies.append(item_proxy_ref)

        def finalize(item_proxy_ref: _WeakReferenceType, item_proxies: typing.List[_WeakReferenceType]) -> None:
            item_proxies.pop(item_proxies.index(item_proxy_ref))

        weakref.finalize(item_proxy, finalize, item_proxy_ref, self.__item_proxies)
        return item_proxy

    def update_item_reference_uuid(self, item_proxy: ItemProxy, item_uuid: typing.Optional[uuid.UUID]) -> None:
        if item_uuid and item_uuid in self.__entity_map:
            entity_ref = self.__entity_map.get(item_uuid)
            if entity_ref:
                item_proxy.item = entity_ref()
                return
        item_proxy.item = None



class Field(abc.ABC):
    """A value-holding field in an entity or another field.

    The context is used to resolve references and must be valid to read entities containing references.
    """
    def __init__(self, context: typing.Optional[EntityContext]) -> None:
        self.__context = context

    def close(self) -> None:
        pass

    @property
    def _context(self) -> typing.Optional[EntityContext]:
        return self.__context

    def set_context(self, context: typing.Optional[EntityContext]) -> None:
        """Set the context.

        The context can be changed from None to a value or a value to None.

        Subclasses should propagate the context to their sub fields.
        """
        assert (self.__context is None) != (context is None), f"{self.__context} {context}"
        self.__context = context

    @abc.abstractmethod
    def read(self, dict_value: typing.Any) -> Field: ...

    @abc.abstractmethod
    def write(self) -> DictValue: ...

    # track the parent entity and relationship for if there is one
    def _set_entity_parent(self, entity_parent: EntityParent) -> None:
        pass

    # not abstract to avoid type checking issues until mypy supports abstract properties
    @property
    def field_value(self) -> typing.Any:
        return None

    # not abstract to avoid type checking issues until mypy supports abstract properties
    def set_field_value(self, container: ItemProxyEntity, value: typing.Any) -> None:
        pass

    def field_by_key(self, key: str) -> Field:
        raise AttributeError()

    def field_by_index(self, index: int) -> Field:
        raise AttributeError()


class PropertyField(Field):
    """A value property field, which can hold a simple value."""

    def __init__(self, context: typing.Optional[EntityContext], type: str, optional: bool, default: typing.Any) -> None:
        super().__init__(context)
        self.__type = type
        self.__optional = optional
        self.__default = default
        self.__value = default

    def read(self, dict_value: typing.Any) -> Field:
        self.__value = build_value(self.__type, dict_value)
        return self

    def write(self) -> DictValue:
        return dict_value(self.__type, self.__value)

    @property
    def field_value(self) -> typing.Any:
        return self.__value

    def set_field_value(self, container: ItemProxyEntity, value: typing.Any) -> None:
        self.__value = value


class TupleField(Field):
    """A tuple property field, which can hold an indefinite tuple of simple values."""

    def __init__(self, context: typing.Optional[EntityContext], type: FieldType, optional: bool, default_values: typing.Optional[typing.Sequence[typing.Any]]) -> None:
        super().__init__(context)
        assert isinstance(type, FieldType)
        assert default_values is None or isinstance(default_values, (tuple, list))
        self.__type = type
        self.__optional = optional
        self.__default_values = list(default_values) if default_values is not None else None
        self.__fields: typing.Optional[typing.Tuple[Field, ...]] = None
        self.read(default_values)

    def close(self) -> None:
        self.__clear()
        super().close()

    def __clear(self) -> None:
        if self.__fields is not None:
            for field in self.__fields:
                field.close()
            self.__fields = None

    def set_context(self, context: typing.Optional[EntityContext]) -> None:
        super().set_context(context)
        if self.__fields:
            for field in self.__fields:
                field.set_context(context)

    def read(self, dict_value: typing.Any) -> Field:
        self.__clear()
        if isinstance(dict_value, (tuple, list)):
            self.__fields = tuple(self.__type.create_and_read(self._context, v) for v in dict_value)
        return self

    def write(self) -> DictValue:
        if self.__fields is not None:
            return list(field.write() for field in self.__fields)
        else:
            return None if self.__optional else copy.copy(self.__default_values)

    @property
    def field_value(self) -> typing.Any:
        return [field.field_value for field in self.__fields] if self.__fields else None

    def set_field_value(self, container: ItemProxyEntity, value: typing.Any) -> None:
        assert isinstance(value, (tuple, list))
        self.__fields = tuple(self.__type.create_and_read(self._context, v) for v in value)


class FixedTupleField(Field):
    """A tuple property field, which can hold a fixed length tuple of simple values."""

    def __init__(self, context: typing.Optional[EntityContext], types: typing.Sequence[FieldType], optional: bool, default_values: typing.Optional[typing.Tuple[typing.Any, ...]]) -> None:
        super().__init__(context)
        assert default_values is None or isinstance(default_values, (tuple, list))
        self.__types = types
        self.__optional = optional
        self.__default_values = list(default_values) if default_values is not None else None
        self.__fields: typing.Optional[typing.Tuple[Field, ...]] = None
        self.read(default_values)

    def close(self) -> None:
        self.__clear()
        super().close()

    def __clear(self) -> None:
        if self.__fields is not None:
            for field in self.__fields:
                field.close()
            self.__fields = None

    def set_context(self, context: typing.Optional[EntityContext]) -> None:
        super().set_context(context)
        if self.__fields:
            for field in self.__fields:
                field.set_context(context)

    def read(self, dict_value: typing.Any) -> Field:
        # TODO: the fields should be fixed
        self.__clear()
        if isinstance(dict_value, (tuple, list)):
            self.__fields = tuple(type.create_and_read(self._context, v) for type, v in zip(self.__types, dict_value))
        return self

    def write(self) -> DictValue:
        if self.__fields is not None:
            return list(field.write() for field in self.__fields)
        else:
            return None if self.__optional else copy.copy(self.__default_values)

    @property
    def field_value(self) -> typing.Any:
        return [field.field_value for field in self.__fields] if self.__fields else None

    def set_field_value(self, container: ItemProxyEntity, value: typing.Any) -> None:
        assert isinstance(value, (tuple, list))
        self.__fields = tuple(type.create_and_read(self._context, v) for type, v in zip(self.__types, value))


class RecordField(Field):
    """A record property field, which can hold a mapping of strings to simple values."""

    def __init__(self, context: typing.Optional[EntityContext], field_type_map: typing.Mapping[str, FieldType]) -> None:
        super().__init__(context)
        self.__field_type_map = field_type_map
        self.__field_map: typing.Dict[str, Field] = dict()
        for k, type in self.__field_type_map.items():
            self.__field_map[k] = type.create(context)

    def close(self) -> None:
        for field in self.__field_map.values():
            field.close()
        self.__field_map = None  # type: ignore
        super().close()

    def set_context(self, context: typing.Optional[EntityContext]) -> None:
        super().set_context(context)
        for field in self.__field_map.values():
            field.set_context(context)

    def read(self, dict_value: typing.Any) -> Field:
        if isinstance(dict_value, (dict)):
            for k, field in self.__field_map.items():
                field.read(dict_value.get(k))
        return self

    def write(self) -> DictValue:
        d = dict()
        for k, field in self.__field_map.items():
            v = field.write()
            if v is not None:
                d[k] = v
        return d

    @property
    def field_value(self) -> typing.Any:
        d = {k: field.field_value for k, field in self.__field_map.items()}
        return collections.namedtuple("record", d.keys())(*d.values())

    def set_field_value(self, container: ItemProxyEntity, value: typing.Any) -> None:
        assert isinstance(value, dict)
        for field in self.__field_map.values():
            field.close()
        for k, field_value in value.items():
            self.__field_map[k] = self.__field_type_map[k].create_and_read(self._context, field_value)


class ArrayField(Field):
    """A array of fields, which can be simple values, references, or components."""

    def __init__(self, context: typing.Optional[EntityContext], type: FieldType, optional: bool) -> None:
        super().__init__(context)
        self.__type = type
        self.__optional = optional
        self.__fields: typing.List[Field] = list()

    def close(self) -> None:
        for field in self.__fields:
            field.close()
        self.__fields = None  # type: ignore
        super().close()

    def set_context(self, context: typing.Optional[EntityContext]) -> None:
        super().set_context(context)
        for field in self.__fields:
            field.set_context(context)

    def read(self, dict_value: typing.Any) -> Field:
        if isinstance(dict_value, (tuple, list)):
            self.__fields = list(self.__type.create_and_read(self._context, item) for item in dict_value)
        else:
            self.__fields = list()
        return self

    def write(self) -> DictValue:
        if self.__fields is not None:
            l = list(field.write() for field in self.__fields)
        else:
            l = list()
        return l if l or not self.__optional else None

    # for an array, the parent of each of the children is the parent of the array
    def _set_entity_parent(self, entity_parent: EntityParent) -> None:
        for entity in typing.cast(typing.Sequence[Entity], self.field_value):
            if entity:
                entity._set_entity_parent(entity_parent)

    @property
    def field_value(self) -> typing.Sequence[typing.Any]:
        return [field.field_value for field in self.__fields]

    def set_field_value(self, container: ItemProxyEntity, value: typing.Any) -> None:
        assert isinstance(value, (tuple, list))
        self.__fields = list(self.__type.create_and_read(self._context, v) for v in value)

    def insert_value(self, container: ItemProxyEntity, index: int, value: typing.Any) -> None:
        if isinstance(value, Entity):
            field = self.__type.create(self._context)
            field.set_field_value(container, value)  # no container yet
            self.__fields.insert(index, field)
        else:
            raise IndexError()

    def remove_value_at_index(self, index: int) -> None:
        field = self.__fields.pop(index)
        field.set_field_value(None, None)  # no container

    def field_by_index(self, index: int) -> Field:
        return self.__fields[index]


class MapField(Field):
    """A map of strings to fields, which can be simple values, references, or components."""

    def __init__(self, context: typing.Optional[EntityContext], key: FieldType, value: FieldType, optional: bool) -> None:
        super().__init__(context)
        self.__key = key
        self.__value = value
        self.__optional = optional
        self.__map: typing.Dict[str, Field] = dict()

    def close(self) -> None:
        for field in self.__map.values():
            field.close()
        self.__map = None  # type: ignore
        super().close()

    def set_context(self, context: typing.Optional[EntityContext]) -> None:
        super().set_context(context)
        for field in self.__map.values():
            field.set_context(context)

    def read(self, dict_value: typing.Any) -> Field:
        if isinstance(dict_value, dict):
            self.__map = {k: self.__value.create_and_read(self._context, v) for k, v in dict_value.items()}
        else:
            self.__map = dict()
        return self

    def write(self) -> DictValue:
        if self.__map is not None:
            d = {k: v.write() for k, v in self.__map.items()}
        else:
            d = dict()
        return d if d or not self.__optional else None

    @property
    def field_value(self) -> typing.Any:
        return {k: field.field_value for k, field in self.__map.items()}

    def set_field_value(self, container: ItemProxyEntity, value: typing.Any) -> None:
        assert isinstance(value, dict)
        self.__fields = {k: self.__value.create_and_read(self._context, v) for k, v in value.items()}

    def set_value(self, container: ItemProxyEntity, key: str, value: typing.Any) -> None:
        if isinstance(value, Entity):
            field = self.__value.create(self._context)
            field.set_field_value(container, value)  # no container yet
            self.__map[key] = field
        else:
            raise IndexError()


class ReferenceField(Field):
    """A reference field, references another entity without cascading delete."""

    def __init__(self, context: typing.Optional[EntityContext], type: EntityType) -> None:
        super().__init__(context)
        self.__type = type
        self.__reference_uuid: typing.Optional[uuid.UUID] = None
        self.__reference: typing.Optional[ItemProxy] = None  # proxy is only valid when context is valid
        self.__shadow_item: typing.Optional[Entity] = None  # used when proxy is None
        if context:
            super().set_context(None)  # unset it safely
            self.set_context(context)

    def close(self) -> None:
        if self.__reference:
            self.__reference.close()
            self.__reference = None
        super().close()

    def set_context(self, context: typing.Optional[EntityContext]) -> None:
        super().set_context(context)
        if self._context:
            if not self.__reference:
                self.__reference = self._context.create_item_reference(self.__reference_uuid if self.__reference_uuid else None)
            if self.__shadow_item:
                self.__reference.item = self.__shadow_item
                self.__shadow_item = None
        elif self.__reference:
            self.__shadow_item = self.__reference.item
            self.__reference.close()
            self.__reference = None

    def read(self, dict_value: typing.Any) -> Field:
        if False and isinstance(dict_value, dict):
            # NOTE: disabled until needed, use direct untyped references for now in all cases.
            # support typed references which will have a type and a uuid.
            # type is unchecked, for now.
            self.__reference_uuid = uuid.UUID(dict_value["uuid"])
        else:
            # support direct untyped uuid references.
            self.__reference_uuid = uuid.UUID(dict_value)
        if self.__reference and self._context:
            self._context.update_item_reference_uuid(self.__reference, self.__reference_uuid)
        return self

    def write(self) -> DictValue:
        return str(self.__reference_uuid) if self.__reference_uuid else None

    @property
    def field_value(self) -> typing.Any:
        if self.__reference:
            return self.__reference.item
        else:
            return self.__shadow_item

    def set_field_value(self, container: ItemProxyEntity, value: typing.Any) -> None:
        item = typing.cast(Entity, value)
        if item:
            item_uuid = item.uuid  # prefer _get_field_value; but use direct accessor for legacy Swift PersistentObject compatibility
            self.__reference_uuid = item_uuid
            if self.__reference:
                self.__reference.item = item
            else:
                self.__shadow_item = item
        else:
            if self.__reference:
                self.__reference.item = None
            else:
                self.__shadow_item = None
            self.__reference_uuid = None

    @property
    def reference_uuid(self) -> typing.Optional[uuid.UUID]:
        return self.__reference_uuid

    def _update_reference_uuid(self, reference_uuid: typing.Optional[uuid.UUID]) -> None:
        assert not self.__reference
        assert not self.__shadow_item
        self.__reference_uuid = reference_uuid


class ComponentField(Field):
    """A component field, references another entity with cascading delete."""

    def __init__(self, context: typing.Optional[EntityContext], entity_id: str) -> None:
        super().__init__(context)
        self.__type = get_entity_type(entity_id)
        self.__entity: typing.Optional[Entity] = None

    def close(self) -> None:
        if self.__entity:
            self.__entity.close()
            self.__entity = None
        super().close()

    def set_context(self, context: typing.Optional[EntityContext]) -> None:
        super().set_context(context)
        if self.__entity:
            self.__entity._set_entity_context(context)

    def read(self, dict_value: typing.Any) -> Field:
        assert self.__type
        self.__entity = self.__type.create(self._context, dict_value)
        return self

    def write(self) -> DictValue:
        return self.__entity.write_to_dict() if self.__entity else None

    @property
    def field_value(self) -> typing.Any:
        return self.__entity

    def set_field_value(self, container: ItemProxyEntity, value: typing.Any) -> None:
        assert value is None or not value._container
        if self.__entity:
            self.__entity._container = None
            if self.__entity._entity_context:
                self.__entity._set_entity_context(None)
        self.__entity = value
        if self.__entity:
            self.__entity._container = container
            if self._context:
                self.__entity._set_entity_context(self._context)

    def field_by_key(self, key: str) -> Field:
        assert self.__entity
        return self.__entity.get_field(key)


class ComponentPlaceholderField(Field):
    """Keep the dict, but otherwise be non-functional."""

    def __init__(self, context: typing.Optional[EntityContext]) -> None:
        super().__init__(context)
        self.__d = typing.cast(DictValue, None)

    def read(self, dict_value: typing.Any) -> Field:
        self.__d = copy.deepcopy(dict_value)
        return self

    def write(self) -> DictValue:
        return copy.deepcopy(self.__d)


class FieldType(abc.ABC):
    """A field type that provides the ability to create a field of a given type.

    The field class is a callable returning a field. The args and kwargs are passed to the callable.
    """
    def __init__(self, field_class: typing.Callable[..., Field], *args: typing.Any, **kwargs: typing.Any) -> None:
        assert callable(field_class)
        self.__field_class = field_class
        self.__args = args
        self.__kwargs = kwargs

    def __repr__(self) -> str:
        return self._get_repr([])

    def _get_repr(self, parents: typing.List[typing.Any]) -> str:
        raise NotImplementedError()

    def visit(self, value: typing.Any, accessor: Accessor, visitor: Visitor) -> None:
        raise NotImplementedError()

    @property
    def _field_class(self) -> typing.Callable[..., Field]:
        return self.__field_class

    @property
    def _args(self) -> typing.Tuple[typing.Any, ...]:
        return self.__args

    @property
    def _kwargs(self) -> typing.Dict[str, typing.Any]:
        return self.__kwargs

    def _call(self, context: typing.Optional[EntityContext], field_class: typing.Callable[..., Field], *args: typing.Any, **kwargs: typing.Any) -> Field:
        return field_class(context, *args, **kwargs)

    def create(self, context: typing.Optional[EntityContext]) -> Field:
        return self._call(context, self.__field_class, *self.__args, **self.__kwargs)

    def create_and_read(self, context: typing.Optional[EntityContext], dict_value: DictValue) -> Field:
        return self.create(context).read(dict_value)


class PropertyType(FieldType):
    def __init__(self, type: str, optional: bool, default: typing.Any) -> None:
        super().__init__(PropertyField, type, optional, default)
        self.type = type
        self.optional = optional
        self.default = default

    def _get_repr(self, parents: typing.List[typing.Any]) -> str:
        return self.type

    def visit(self, value: typing.Any, accessor: Accessor, visitor: Visitor) -> None:
        pass


class TupleType(FieldType):
    def __init__(self, type: FieldType, optional: bool, default: typing.Optional[typing.Sequence[typing.Any]]) -> None:
        super().__init__(TupleField, type, optional, default)
        self.type = type
        self.optional = optional

    def _get_repr(self, parents: typing.List[typing.Any]) -> str:
        if self not in parents:
            return f"tuple[{self.type}]"
        else:
            return f"tuple[self]"

    def visit(self, value: typing.Tuple[typing.Any], accessor: Accessor, visitor: Visitor) -> None:
        for index, tuple_value in enumerate(value):
            next_accessor = IndexAccessor(accessor, self.type, index)
            # inform the visitor of the tuple entry
            visitor.visit(next_accessor)
            # then visit the value
            self.type.visit(tuple_value, next_accessor, visitor)


class FixedTupleType(FieldType):
    def __init__(self, types: typing.Sequence[FieldType], optional: bool, default: typing.Optional[typing.Tuple[typing.Any, ...]]) -> None:
        super().__init__(FixedTupleField, types, optional, default)
        self.types = list(types)
        self.optional = optional

    def _get_repr(self, parents: typing.List[typing.Any]) -> str:
        types_str = ", ".join(t._get_repr(parents + [self]) if t not in parents else "self" for t in self.types)
        return f"fixed_tuple[{types_str}]"

    def visit(self, value: typing.Tuple[typing.Any], accessor: Accessor, visitor: Visitor) -> None:
        for index, tuple_value in enumerate(value):
            next_accessor = IndexAccessor(accessor, self.types[index], index)
            # inform the visitor of the tuple entry
            visitor.visit(next_accessor)
            # then visit the value
            self.types[index].visit(tuple_value, next_accessor, visitor)


class RecordType(FieldType):
    def __init__(self, field_type_map: typing.Mapping[str, FieldType]) -> None:
        super().__init__(RecordField, field_type_map)
        self.__field_type_map = dict(field_type_map)

    @property
    def _field_type_map(self) -> typing.Mapping[str, FieldType]:
        return self.__field_type_map

    def _get_repr(self, parents: typing.List[typing.Any]) -> str:
        types_str = ", ".join(f"{t}: {f._get_repr(parents + [self]) if f not in parents else 'self'}" for t, f in self.__field_type_map.items())
        return f"record[{types_str}]"

    def visit(self, value: typing.Any, accessor: Accessor, visitor: Visitor) -> None:
        for field_name, field_type in self.__field_type_map.items():
            next_accessor = FieldAccessor(accessor, field_type, field_name)
            # inform the visitor of the field
            visitor.visit(next_accessor)
            # then visit each field
            field_value = getattr(value, field_name)
            if field_value is not None:
                field_type.visit(field_value, next_accessor, visitor)


class ArrayType(FieldType):
    def __init__(self, type: FieldType, optional: bool) -> None:
        super().__init__(ArrayField, type, optional)
        self.type = type
        self.optional = optional

    def _get_repr(self, parents: typing.List[typing.Any]) -> str:
        if self not in parents:
            return f"array[{self.type._get_repr(parents + [self])}]"
        else:
            return f"array[self]"

    def visit(self, value: typing.Sequence[typing.Any], accessor: Accessor, visitor: Visitor) -> None:
        for index, array_value in enumerate(value):
            next_accessor = IndexAccessor(accessor, self.type, index)
            # inform the visitor of the tuple entry
            visitor.visit(next_accessor)
            # then visit the value
            self.type.visit(array_value, next_accessor, visitor)


class MapType(FieldType):
    def __init__(self, key: str, value_type: FieldType, optional: bool) -> None:
        super().__init__(MapField, key, value_type, optional)
        self.key = key
        self.value_type = value_type
        self.optional = optional

    def _get_repr(self, parents: typing.List[typing.Any]) -> str:
        if self not in parents:
            return f"map[{self.key}: {self.value_type._get_repr(parents + [self])}]"
        else:
            return f"map[{self.key}: self]"

    def visit(self, value: typing.Mapping[typing.Any, typing.Any], accessor: Accessor, visitor: Visitor) -> None:
        for map_key, map_value in value.items():
            next_accessor = MapAccessor(accessor, self.value_type, map_key)
            # inform the visitor of the tuple entry
            visitor.visit(next_accessor)
            # then visit the value
            self.value_type.visit(map_value, next_accessor, visitor)


class ReferenceType(FieldType):
    def __init__(self, type: typing.Optional[EntityType]) -> None:
        super().__init__(ReferenceField, type)
        self.type = type

    def _get_repr(self, parents: typing.List[typing.Any]) -> str:
        if self not in parents:
            return f"reference[{self.type._get_repr(parents + [self]) if self.type else 'None'}]"
        else:
            return f"reference[self]"

    def visit(self, value: typing.Any, accessor: Accessor, visitor: Visitor) -> None:
        pass


class ComponentType(FieldType):
    def __init__(self, entity_id: str, required: bool) -> None:
        super().__init__(ComponentField, entity_id)
        self.entity_id = entity_id
        self.required = required

    def create_and_read(self, context: typing.Optional[EntityContext], dict_value: DictValue) -> Field:
        assert isinstance(dict_value, dict)
        d = dict_value
        entity_type = get_entity_type(d["type"])
        if entity_type:
            field_type = entity_type.entity_id if entity_type else self._args[0]
            return self._call(context, self._field_class, *((field_type, ) + self._args[1:]), **self._kwargs).read(dict_value)
        else:
            return ComponentPlaceholderField(context).read(dict_value)

    def _get_repr(self, parents: typing.List[typing.Any]) -> str:
        if self not in parents:
            return f"component[{entity_types[self.entity_id]._get_repr(parents + [self])}]"
        else:
            return f"component[self]"

    def visit(self, value: Entity, accessor: Accessor, visitor: Visitor) -> None:
        entity_types[self.entity_id].visit(value, accessor, visitor)


_EntityTransform = typing.Callable[[PersistentDictType], PersistentDictType]
_EntityTransforms = typing.Tuple[_EntityTransform, _EntityTransform]


class EntityParent:
    # this is for compatibility with the persistence framework and should match the corresponding class there.

    def __init__(self, parent: Entity, relationship_name: typing.Optional[str] = None, item_name: typing.Optional[str] = None) -> None:
        self.__weak_parent = weakref.ref(parent)
        self.relationship_name = relationship_name
        self.item_name = item_name

    @property
    def parent(self) -> typing.Optional[Entity]:
        return self.__weak_parent()


class Entity(Observable.Observable):
    """An instance of an entity type.

    To support deepcopy, subclasses must be able to be initialized with no arguments (passing entity_type to this init)
    or override deepcopy.
    """
    def __init__(self, entity_type: typing.Optional[EntityType] = None, context: typing.Optional[EntityContext] = None) -> None:
        super().__init__()
        assert entity_type, str(self.__class__)
        self.__context: typing.Optional[EntityContext] = None
        self.__parent: typing.Optional[EntityParent] = None
        self.__entity_type = entity_type
        self.__version = entity_type._version
        self.__field_type_map = entity_type._field_type_map
        self.__field_dict : typing.Dict[str, Field] = dict()
        self.__renames = entity_type._renames
        self.__transforms = entity_type._transforms
        self._container = None
        for field_name, field_type in self.__field_type_map.items():
            self.__field_dict[field_name] = field_type.create(self.__context)
        self._set_field_value("uuid", uuid.uuid4())
        self._set_field_value("modified", DateTime.utcnow())
        if context:
            self._set_entity_context(context)

    def close(self) -> None:
        for field in self.__field_dict.values():
            field.close()
        self.__field_dict.clear()

    def __deepcopy__(self, memo: typing.Dict[typing.Any, typing.Any]) -> Entity:
        entity_copy = self._deepcopy()
        memo[id(self)] = entity_copy
        return entity_copy

    def __repr__(self) -> str:
        return f"entity '{self.__entity_type.entity_id}' at 0x{id(self):x}"

    def _create(self, context: typing.Optional[EntityContext]) -> Entity:
        return self.__class__(self.__entity_type, context)

    def _deepcopy(self) -> Entity:
        context = SimpleEntityContext()
        entity_copy = self._create(context)
        entity_copy.read(self.write_to_dict())
        entity_copy._set_entity_context(None)
        return entity_copy

    @property
    def _entity_context(self) -> typing.Optional[EntityContext]:
        return self.__context

    def _set_entity_context(self, context: typing.Optional[EntityContext]) -> None:
        assert (self.__context is None) != (context is None), f"{self.__context} {context}"
        if self.__context:
            self.__context.unregister(self)
        self.__context = context
        if self.__context:
            self.__context.register(self)
        for field in self.__field_dict.values():
            field.set_context(context)

    @property
    def _entity_parent(self) -> typing.Optional[EntityParent]:
        return self.__parent

    def _set_entity_parent(self, parent: typing.Optional[EntityParent]) -> None:
        self.__parent = parent

    @property
    def uuid(self) -> uuid.UUID:
        return typing.cast(uuid.UUID, self._get_field_value("uuid"))

    @uuid.setter
    def uuid(self, value: uuid.UUID) -> None:
        self._set_field_value("uuid", value)

    @property
    def modified(self) -> datetime.datetime:
        return typing.cast(datetime.datetime, self._get_field_value("modified"))

    def read(self, properties: PersistentMappingType) -> Entity:
        # unregister old uuid
        if self.__context:
            self.__context.unregister(self)
        properties = self.__transforms[0](dict(properties))  # transform forward
        for field_name, field_type in self.__field_type_map.items():
            d = properties.get(self.__renames.get(field_name, field_name))
            if d is not None:
                field = self.__field_dict[field_name]
                field.read(d)
                field._set_entity_parent(EntityParent(self, relationship_name=field_name))
        # register new uuid
        if self.__context:
            self.__context.register(self)
        return self

    def write(self) -> DictValue:
        return self.write_to_dict()

    def write_to_dict(self) -> PersistentDictType:
        d: typing.Dict[str, DictValue] = dict()
        d["type"] = self.__entity_type.entity_id
        if self.__version:
            d["version"] = self.__version
        for field_name, field in self.__field_dict.items():
            dd = field.write()
            if dd is not None:
                d[self.__renames.get(field_name, field_name)] = dd
        return self.__transforms[1](d)  # transform back

    @property
    def entity_type(self) -> EntityType:
        return self.__entity_type

    def _set_modified(self, modified: datetime.datetime) -> None:
        field = self.__get_field("modified")
        if field:
            field.set_field_value(self, modified)
        if self._container:
            self._container._set_modified(modified)

    def __getattr__(self, name: str) -> typing.Any:
        if name in self.__dict__.get("_Entity__field_type_map", dict()):
            return self._get_field_value(name)
        raise AttributeError(f"Unknown attribute {name}")

    def __setattr__(self, name: str, value: typing.Any) -> None:
        field_dict = self.__dict__.get("_Entity__field_dict", dict())
        if name in field_dict:
            self._set_field_value(name, value)
        else:
            super().__setattr__(name, value)

    def __get_field(self, name: str) -> typing.Optional[Field]:
        return self.__field_dict.get(name)

    def get_field(self, name: str) -> Field:
        return typing.cast(Field, self.__get_field(name))

    def _get_field_value(self, name: str) -> typing.Any:
        field = self.__get_field(name)
        if field:
            return field.field_value
        raise AttributeError()

    def _set_field_value(self, name: str, value: typing.Any) -> None:
        field = self.__get_field(name)
        if field:
            field.set_field_value(self, value)
            self._field_value_changed(name, field.write())
            if name not in ("modified", "uuid"):
                self._set_modified(DateTime.utcnow())
            self.property_changed_event.fire(name)
        else:
            raise AttributeError()

    def _get_array_item(self, name: str, index: int) -> typing.Any:
        array_field = typing.cast(typing.Optional[ArrayField], self.__get_field(name))
        if array_field:
            return array_field.field_value[index]
        else:
            raise AttributeError()

    def _get_array_items(self, name: str) -> typing.Sequence[typing.Any]:
        array_field = typing.cast(typing.Optional[ArrayField], self.__get_field(name))
        if array_field:
            return array_field.field_value
        else:
            raise AttributeError()

    def _insert_item(self, name: str, index: int, item: ItemProxyEntity) -> None:
        array_field = typing.cast(typing.Optional[ArrayField], self.__get_field(name))
        if array_field:
            entity = typing.cast(Entity, item)
            entity._set_entity_parent(EntityParent(self, relationship_name=name))
            array_field.insert_value(self, index, item)  # passing self for container
            self._set_modified(DateTime.utcnow())
            self._item_inserted(name, index, item)
            self.item_inserted_event.fire(name, item, index)
        else:
            raise AttributeError()

    def _append_item(self, name: str, item: ItemProxyEntity) -> None:
        self._insert_item(name, len(self._get_field_value(name)), item)

    def _remove_item(self, name: str, item: ItemProxyEntity) -> None:
        array_field = typing.cast(typing.Optional[ArrayField], self.__get_field(name))
        if array_field:
            index = list(self._get_array_items(name)).index(item)
            array_field.remove_value_at_index(index)  # passing self for container
            entity = typing.cast(Entity, item)
            entity._set_entity_parent(None)
            self._set_modified(DateTime.utcnow())
            self._item_removed(name, index, item)
            self.item_removed_event.fire(name, item, index)
        else:
            raise AttributeError()

    def _remove_all_items(self, name: str) -> None:
        array_field = typing.cast(typing.Optional[ArrayField], self.__get_field(name))
        if array_field:
            for item in list(self._get_array_items(name)):
                self._remove_item(name, item)

    # these methods can be overridden to provide custom behavior for storage and notification

    def _field_value_changed(self, name: str, value: typing.Any) -> None:
        pass

    def _item_inserted(self, name: str, index: int, item: ItemProxyEntity) -> None:
        pass

    def _item_removed(self, name: str, index: int, item: ItemProxyEntity) -> None:
        pass

    # compatibility functions for persistent object

    def begin_reading(self) -> None:
        pass

    def read_from_dict(self, properties: PersistentMappingType) -> None:
        self.read(properties)

    def finish_reading(self) -> None:
        pass

    def about_to_be_inserted(self, container: typing.Any) -> None:
        self._container = container

    def about_to_be_removed(self, container: typing.Any) -> None:
        self._container = None

    @property
    def container(self) -> typing.Any:
        return self._container

    @property
    def persistent_object_context(self) -> typing.Any:
        return self._entity_context

    @persistent_object_context.setter
    def persistent_object_context(self, value: typing.Any) -> None:
        self._set_entity_context(value)

    @property
    def persistent_object_parent(self) -> typing.Any:
        return self._entity_parent

    @persistent_object_parent.setter
    def persistent_object_parent(self, value: typing.Any) -> typing.Any:
        self._set_entity_parent(value)

    @property
    def item_names(self) -> typing.List[str]:
        return [k for k, f in self.__field_type_map.items() if isinstance(f, ComponentType)]

    @property
    def relationship_names(self) -> typing.List[str]:
        return [k for k, f in self.__field_type_map.items() if isinstance(f, ArrayType)]

    def get_relationship_items(self, name: str) -> typing.List[typing.Any]:
        return typing.cast(typing.List[typing.Any], getattr(self, name))

    def item_specifier(self) -> typing.Any:
        class ItemSpecifier:

            def __init__(self, *, item: typing.Optional[Entity] = None, item_uuid: typing.Optional[uuid.UUID] = None) -> None:
                self.__item_uuid = item.uuid if item else item_uuid
                assert (self.__item_uuid is None) or isinstance(self.__item_uuid, uuid.UUID)

            def __hash__(self) -> typing.Any:
                return hash(self.__item_uuid)

            def __eq__(self, other: typing.Any) -> bool:
                if isinstance(other, self.__class__):
                    return bool(self.__item_uuid == other.__item_uuid)
                return False

            @property
            def item_uuid(self) -> typing.Optional[uuid.UUID]:
                return self.__item_uuid

            def write(self) -> typing.Optional[DictValue]:
                if self.__item_uuid:
                    return str(self.__item_uuid)
                return None

            @staticmethod
            def read(d: DictValue) -> typing.Optional[ItemSpecifier]:
                if isinstance(d, str):
                    return ItemSpecifier(item_uuid=uuid.UUID(d))
                elif isinstance(d, uuid.UUID):
                    return ItemSpecifier(item_uuid=d)
                elif isinstance(d, dict) and "item_uuid" in d:
                    return ItemSpecifier(item_uuid=uuid.UUID(d["item_uuid"]))
                elif isinstance(d, dict) and "uuid" in d:
                    return ItemSpecifier(item_uuid=uuid.UUID(d["uuid"]))
                else:
                    return None

        return ItemSpecifier(item=self)


EntityAttributePropertyType = typing.TypeVar('EntityAttributePropertyType')

class EntityAttribute(typing.Generic[EntityAttributePropertyType]):
    def __set_name__(self, owner: Entity, name: str) -> None:
        self.name = name

    def __get__(self, obj: Entity, type: typing.Optional[typing.Type[Entity]] = None) -> EntityAttributePropertyType:
        return typing.cast(EntityAttributePropertyType, obj._get_field_value(self.name))

    def __set__(self, obj: Entity, value: EntityAttributePropertyType) -> None:
        obj._set_field_value(self.name, value)


def no_transform(x: PersistentDictType) -> PersistentDictType:
    return x


class EntityType:
    def __init__(self, entity_id: str, base: typing.Optional[EntityType], version: typing.Optional[int], field_type_map: typing.Mapping[str, FieldType], factory: typing.Optional[typing.Callable[[EntityType, typing.Optional[EntityContext]], Entity]] = None) -> None:
        self.__entity_id = entity_id
        self.__base = base
        self.__version = version
        self.__renames: typing.Dict[str, str] = dict()
        self.__transforms: _EntityTransforms = (no_transform, no_transform)
        self.__field_type_map: typing.Dict[str, FieldType] = dict()
        self.__field_type_map["uuid"] = prop(UUID)
        self.__field_type_map["modified"] = prop(TIMESTAMP)
        self.__factory = factory
        while base:
            for k, v in base._field_type_map.items():
                if not k in ("uuid", "modified"):
                    self.__field_type_map[k] = v
            for k2, v2 in base._renames.items():  # use '2' versions to satisfy mypy
                self.__renames[k2] = v2
            base = base.base
        for k, v in field_type_map.items():
            self.__field_type_map[k] = v
        register_entity_type(entity_id, self)

    def __repr__(self) -> str:
        return self._get_repr([])

    def _get_repr(self, parents: typing.List[typing.Any]) -> str:
        version_str = f", {self.__version}" if self.__version is not None else str()
        base_str = f" ({self.__base.entity_id})" if self.__base else str()
        fields_str = f" (" + ", ".join([f"{n}: {t._get_repr(parents + [self]) if t not in parents else 'self'}" for n, t in self.__field_type_map.items()]) + ")"
        r = f"entity [{self.__entity_id}{version_str}]{base_str}{fields_str}"
        return r

    def visit(self, value: Entity, accessor: Accessor, visitor: Visitor) -> None:
        for field_name, field_type in self.__field_type_map.items():
            next_accessor = FieldAccessor(accessor, field_type, field_name)
            # inform the visitor of the field
            visitor.visit(next_accessor)
            # then visit each field
            field_value = getattr(value, field_name, None)
            if field_value is not None:
                field_type.visit(field_value, next_accessor, visitor)

    @property
    def _version(self) -> typing.Optional[int]:
        return self.__version

    @property
    def _field_type_map(self) -> typing.Mapping[str, FieldType]:
        return self.__field_type_map

    @property
    def _renames(self) -> typing.Mapping[str, str]:
        return self.__renames

    @property
    def _transforms(self) -> _EntityTransforms:
        return self.__transforms

    def create(self, context: typing.Optional[EntityContext] = None, d: typing.Optional[PersistentMappingType] = None) -> Entity:
        type = d.get("type", None) if d else None
        if type:
            concrete_subclasses = self.concrete_subclasses
            if concrete_subclasses:
                for concrete_subclass in concrete_subclasses:
                    if type == concrete_subclass.entity_id:
                        return concrete_subclass.create(context, d)
        # figure out which factory to use. if the factory is set when constructing this type, use it.
        # otherwise, look for a registered factory in this class and each base class. otherwise, use
        # a generic Entity.
        factory: typing.Optional[typing.Callable[[EntityType, typing.Optional[EntityContext]], typing.Optional[Entity]]] = self.__factory
        entity: typing.Optional[Entity] = None
        base_entity: typing.Optional[EntityType] = self
        while not factory and base_entity:
            global _entity_factories
            factory = _entity_factories.get(base_entity.entity_id)
            if factory:
                break
            base_entity = base_entity.base
        if factory:
            entity = factory(self, context)
        if not entity:
            entity = Entity(self, context)
        # now read the entity from the dict.
        if d is not None:
            entity.read(d)
        return entity

    @property
    def base(self) -> typing.Optional[EntityType]:
        return self.__base

    @property
    def entity_id(self) -> str:
        return self.__entity_id

    def rename(self, field_name: str, storage_field_name: str) -> None:
        self.__renames[field_name] = storage_field_name

    def transform(self, forward: _EntityTransform, backward: _EntityTransform) -> None:
        self.__transforms = (forward, backward)

    def is_subclass_of(self, entity_type: EntityType) -> bool:
        return self.base is not None and (self.base == entity_type or self.base.is_subclass_of(entity_type))

    @property
    def subclasses(self) -> typing.List[EntityType]:
        return [entity_type for entity_type in entity_types.values() if entity_type.is_subclass_of(self)]

    @property
    def concrete_subclasses(self) -> typing.List[EntityType]:
        return [entity_type for entity_type in entity_types.values() if entity_type.is_subclass_of(self) and not entity_type.subclasses]


def prop(type: str, optional: bool = False, *, default: typing.Any = None) -> PropertyType:
    return PropertyType(type, optional, default)

def indefinite_tuple(type: FieldType, optional: bool = False, default: typing.Optional[typing.Sequence[typing.Any]] = None) -> TupleType:
    return TupleType(type, optional, default)

def fixed_tuple(types: typing.List[FieldType], optional: bool = False, default: typing.Optional[typing.Tuple[typing.Any, ...]] = None) -> FixedTupleType:
    return FixedTupleType(types, optional, default)

def record(field_type_map: typing.Dict[str, FieldType]) -> RecordType:
    return RecordType(field_type_map)

def array(type: FieldType, optional: bool = False) -> ArrayType:
    return ArrayType(type, optional)

def map(key: str, value: FieldType, optional: bool = False) -> MapType:
    return MapType(key, value, optional)

def reference(type: typing.Optional[EntityType] = None) -> ReferenceType:
    return ReferenceType(type)

def component(type: typing.Union[EntityType, str], required: bool = True) -> ComponentType:
    if isinstance(type, EntityType):
        return ComponentType(type.entity_id, required)
    else:  # str, used for forward or self references
        return ComponentType(type, required)

def entity(entity_id: str, base: typing.Optional[EntityType], version: typing.Optional[int], field_type_map: typing.Dict[str, FieldType], factory: typing.Optional[typing.Callable[[EntityType, typing.Optional[EntityContext]], Entity]] = None) -> EntityType:
    return EntityType(entity_id, base, version, field_type_map, factory)

def read_json(path: pathlib.Path) -> PersistentDictType:
    """Read the json file from path. Rename path if not readable."""
    properties = dict()
    if path and path.exists():
        try:
            with path.open("r") as fp:
                properties = json.load(fp)
        except Exception:
            os.replace(path, path.with_suffix(".bak"))
    return properties

def read_entity(entity_type: EntityType, context: EntityContext, path: pathlib.Path, o: str) -> Entity:
    """Read the entity from the path."""
    properties = read_json(path)
    return entity_type.create(context, properties)


_entity_factories = dict[str, typing.Callable[[EntityType, typing.Optional[EntityContext]], typing.Optional[Entity]]]()

# this is required to be able to construct a custom class from the entity. the old technique of passing the factory
# into the entity type is not sufficient because the custom class would have to be known at the time the entity type
# is constructed. this is not always possible when the model is declared independently of the custom classes.
def register_entity_factory(entity_type: EntityType, factory: typing.Callable[[EntityType, typing.Optional[EntityContext]], typing.Optional[Entity]]) -> None:
    _entity_factories[entity_type.entity_id] = factory



from __future__ import annotations

# standard libraries
import dataclasses
import enum
import gettext
import logging
import os
import pathlib
import platform
import re
import subprocess
import traceback
import typing
import unicodedata

# third party libraries
# None

# local libraries
from nion.swift.model import ImportExportManager
from nion.swift.model import Utility
from nion.swift import DocumentController
from nion.swift import DisplayPanel
from nion.ui import Declarative
from nion.ui import Dialog
from nion.ui import UserInterface
from nion.utils import Converter
from nion.utils import Geometry
from nion.utils import Model
from nion.utils import Observable

if typing.TYPE_CHECKING:
    from nion.swift.model import DisplayItem

_ = gettext.gettext


class ExportDialogViewModel:

    def __init__(self, title: bool, date: bool, dimensions: bool, sequence: bool, writer: typing.Optional[ImportExportManager.ImportExportHandler], prefix: str = "", directory: str = ""):
        self.include_title = Model.PropertyModel(title)
        self.include_date = Model.PropertyModel(date)
        self.include_dimensions = Model.PropertyModel(dimensions)
        self.include_sequence = Model.PropertyModel(sequence)
        self.include_prefix = Model.PropertyModel(prefix is not None)
        self.prefix = Model.PropertyModel(prefix)
        self.directory = Model.PropertyModel(directory)
        self.writer = Model.PropertyModel(writer)


class ExportDialog(Declarative.Handler):

    def __init__(self, ui: UserInterface.UserInterface, document_controller: DocumentController.DocumentController, display_items: typing.Sequence[DisplayItem.DisplayItem]):
        super().__init__()

        self.ui = ui
        self.__document_controller = document_controller

        # configure the writers. ensure that the writer is set to the last used writer or the first writer if no last used.
        self.__writers = ImportExportManager.ImportExportManager().get_writers()
        io_handler_id = self.ui.get_persistent_string("export_io_handler_id", "png-io-handler")
        writer = ImportExportManager.ImportExportManager().get_writer_by_id(io_handler_id) or self.__writers[0]
        self.writer_index = self.__writers.index(writer)

        # set up directory needed for the viewmodel
        directory = self.ui.get_persistent_string("export_directory", self.ui.get_document_location())

        # the viewmodel is the data model for the dialog. no other variables are needed.
        self.viewmodel = ExportDialogViewModel(True, True, True, True, writer, "", directory)

        # build the UI
        u = Declarative.DeclarativeUI()
        self._build_ui(u)

        # perform export, but save the last used writer
        def handle_export_clicked() -> bool:
            selected_writer = self.viewmodel.writer.value
            writer_id = selected_writer.io_handler_id if selected_writer else "png-io-handler"
            self.export_clicked(display_items, self.viewmodel, ui, document_controller)
            self.ui.set_persistent_string("export_io_handler_id", writer_id)
            return True

        # create the dialog and show it.
        export_text = _("Export")
        items_text = _("Items") if len(display_items) != 1 else _("Item")
        title_text = f"{export_text} ({len(display_items)} {items_text})"
        dialog = typing.cast(Dialog.ActionDialog, Declarative.construct(document_controller.ui, document_controller, u.create_modeless_dialog(self.ui_view, title=title_text), self))
        dialog.add_button(_("Cancel"), self.cancel)
        dialog.add_button(_("Export"), handle_export_clicked)
        dialog.show()

    def choose_directory(self, widget: Declarative.UIWidget) -> None:
        directory = self.viewmodel.directory.value or str()
        selected_directory, directory = self.ui.get_existing_directory_dialog(_("Choose Export Directory"), directory)
        if selected_directory:
            self.viewmodel.directory.value = selected_directory
            self.ui.set_persistent_string("export_directory", selected_directory)

    def on_writer_changed(self, widget: Declarative.UIWidget, current_index: int) -> None:
        writer = self.__writers[current_index]
        self.viewmodel.writer.value = writer

    def _build_ui(self, u: Declarative.DeclarativeUI) -> None:
        writers_names = [getattr(writer, "name") for writer in self.__writers]

        # Export Folder
        directory_label = u.create_row(u.create_label(text="Location:", font='bold'))
        directory_text = u.create_row(u.create_column(u.create_label(text=f"@binding(viewmodel.directory.value)", min_width=280, height=48, word_wrap=True, size_policy_horizontal='min-expanding', text_alignment_vertical='top')))
        self.directory_text_label = directory_text
        directory_button = u.create_row(u.create_push_button(text=_("Select Path..."), on_clicked="choose_directory"), u.create_stretch())

        # Filename
        filename_label = u.create_row(u.create_label(text="Filename:", font='bold'), u.create_stretch())

        # Title
        title_checkbox = u.create_row(
            u.create_check_box(text="Include Title", checked=f"@binding(viewmodel.include_title.value)"), u.create_stretch())

        # Date
        date_checkbox = u.create_row(
            u.create_check_box(text="Include Date", checked=f"@binding(viewmodel.include_date.value)"), u.create_stretch())

        # Dimensions
        dimension_checkbox = u.create_row(
            u.create_check_box(text="Include Dimensions", checked=f"@binding(viewmodel.include_dimensions.value)"), u.create_stretch())

        # Sequence Number
        sequence_checkbox = u.create_row(
            u.create_check_box(text="Include Sequence Number", checked=f"@binding(viewmodel.include_sequence.value)"), u.create_stretch())

        # Prefix
        prefix_label = u.create_label(text=_("Prefix:"))
        prefix_textbox = u.create_line_edit(text=f"@binding(viewmodel.prefix.value)", placeholder_text=_("None"), width=230)
        prefix_row = u.create_row(prefix_label, prefix_textbox, u.create_stretch(), spacing=10)

        # File Type
        file_type_combobox = u.create_combo_box(
            items=writers_names,
            current_index=f"@binding(writer_index)",
            on_current_index_changed="on_writer_changed")
        file_type_label = u.create_label(text=_("File Format:"), font='bold')
        file_type_row = u.create_row(file_type_combobox, u.create_stretch())

        # Build final ui column
        column = u.create_column(directory_label,
                                 directory_text,
                                 directory_button,
                                 filename_label,
                                 prefix_row,
                                 title_checkbox,
                                 date_checkbox,
                                 dimension_checkbox,
                                 sequence_checkbox,
                                 file_type_label,
                                 file_type_row,
                                 spacing=12, margin=12)
        self.ui_view = column

    @staticmethod
    def build_filepath(components: typing.List[str], extension: str, directory_path: pathlib.Path) -> pathlib.Path:
        assert directory_path.is_dir()

        # if extension doesn't start with a '.', add one, so we always know it is there
        if not extension.startswith('.'):
            extension = '.' + extension

        # stick components together for the first part of the filename, underscore delimited excluding blank component
        filename = "_".join(s for s in components if s)
        filename.replace(".", "_")

        filename = Utility.simplify_filename(str(pathlib.Path(filename).with_suffix(extension)))

        # check to see if filename is available, if so return that
        test_filepath = directory_path / pathlib.Path(filename)
        if not test_filepath.exists():
            return test_filepath

        # file must already exist
        next_index = 1
        max_index = 9999
        last_test_filepath: typing.Optional[pathlib.Path] = None
        while next_index <= max_index:
            filename_stem = pathlib.Path(filename).stem
            test_filepath = directory_path / pathlib.Path(f"{filename_stem} {next_index}").with_suffix(extension)
            if not test_filepath.exists():
                return test_filepath
            if test_filepath == last_test_filepath:
                break
            last_test_filepath = test_filepath  # in case we have a bug
            next_index = next_index + 1

        # We have no option here but to just go with the overwrite, either we ran out of index options or had none to begin with
        print(f"Warning: Overwriting file {test_filepath}")
        return test_filepath

    @staticmethod
    def export_clicked(display_items: typing.Sequence[DisplayItem.DisplayItem], viewmodel: ExportDialogViewModel,
                       ui: UserInterface.UserInterface, document_controller: DocumentController.DocumentController) -> None:
        directory_path = pathlib.Path(viewmodel.directory.value or str())
        writer_model = viewmodel.writer
        writer = writer_model.value
        if directory_path.is_dir() and writer:
            export_results = list()
            for index, display_item in enumerate(display_items):
                data_item = display_item.data_item
                file_name: str = ''
                try:
                    components = list()
                    if viewmodel.prefix.value is not None and viewmodel.prefix.value != '':
                        components.append(str(viewmodel.prefix.value))
                    if viewmodel.include_title.value:
                        title = unicodedata.normalize('NFKC', display_item.displayed_title)
                        title = re.sub(r'[^\w\s-]', '', title, flags=re.U).strip()
                        title = re.sub(r'[-\s]+', '-', title, flags=re.U)
                        components.append(title)
                    if viewmodel.include_date.value:
                        # prefer the data item created date, but fall back to the display item created date.
                        created_local = data_item.created_local if data_item else display_item.created_local
                        components.append(created_local.isoformat().replace(':', '').replace('.', '_'))
                    if viewmodel.include_dimensions.value and data_item:
                        components.append("x".join([str(shape_n) for shape_n in data_item.dimensional_shape]))
                    if viewmodel.include_sequence.value:
                        components.append(str(index))
                    filepath = ExportDialog.build_filepath(components, writer.extensions[0], directory_path=directory_path)
                    file_extension = filepath.suffix[1:].lower()
                    if writer.can_write_display_item(display_item, file_extension):
                        ImportExportManager.ImportExportManager().write_display_item_with_writer(writer, display_item, filepath)
                        export_results.append(ExportResult(display_item.displayed_title))
                    else:
                        error_message = _("Cannot export this data to file format")
                        export_results.append(ExportResult(display_item.displayed_title, f"{error_message} {writer.name}"))
                except Exception as e:
                    logging.debug("Could not export image %s / %s", str(data_item), str(e))
                    traceback.print_exc()
                    traceback.print_stack()
                    export_results.append(ExportResult(file_name, str(e)))

            ExportResultDialog(ui, document_controller, export_results, directory_path)

    def cancel(self) -> bool:
        return True


class UnitDescription:
    """A description of a unit. Read only."""

    def __init__(self, unit_id: str, title: str, conversion_factor: float) -> None:
        self.__unit_id = unit_id
        self.__title = title
        self.__conversion_factor = conversion_factor

    @property
    def unit_id(self) -> str:
        return self.__unit_id

    @property
    def title(self) -> str:
        return self.__title

    @property
    def conversion_factor(self) -> float:
        return self.__conversion_factor

    def convert_value_to_pixels(self, value: float) -> float:
        return value * self.__conversion_factor

    def convert_value_from_pixels(self, value: float) -> float:
        return value / self.__conversion_factor


class Quantity:
    """A quantity with a value and a unit description. Read only."""

    def __init__(self, value: float, unit_description: UnitDescription) -> None:
        self.__value = value
        self.__unit_description = unit_description

    def __str__(self) -> str:
        return f"{self.__value} {self.__unit_description.title}"

    def with_value(self, value: float) -> Quantity:
        return Quantity(value, self.__unit_description)

    def with_unit_description(self, unit_description: UnitDescription) -> Quantity:
        return Quantity(unit_description.convert_value_from_pixels(self.pixels), unit_description)

    @property
    def value(self) -> float:
        return self.__value

    @property
    def is_pixels(self) -> bool:
        return self.__unit_description.unit_id == "pixels"

    @property
    def pixels(self) -> float:
        return self.__value * self.__unit_description.conversion_factor


def get_quantity_from_pixels(pixels: float, unit_description: UnitDescription) -> Quantity:
    """Return a quantity from pixels and a unit description."""
    return Quantity(unit_description.convert_value_from_pixels(pixels), unit_description)


# define the units. the conversion factor is the number of pixels per unit.
pixel_unit_description = UnitDescription("pixels", _("Pixels"), 1.0)
inch_unit_description = UnitDescription("inches", _("Inches"), 96.0)
centimeter_unit_description = UnitDescription("centimeters", _("Centimeters"), 37.795275591)


# the list of available units for SVG export
svg_export_unit_descriptions = [
    pixel_unit_description,
    inch_unit_description,
    centimeter_unit_description
]


def calculate_display_size_in_pixels(display_item: DisplayItem.DisplayItem) -> Geometry.IntSize:
    """Return the display size in pixels.

    If the display item is a line plot, use the hardcoded value of 5x3 inches.
    """
    if display_item.display_data_shape and display_item.used_display_type != "line_plot":
        return Geometry.IntSize(height=display_item.display_data_shape[-2], width=display_item.display_data_shape[-1])
    return Geometry.IntSize(
        height=round(inch_unit_description.convert_value_to_pixels(3.0)),
        width=round(inch_unit_description.convert_value_to_pixels(5.0))
    )


class ExportSizeModel(Observable.Observable):
    def __init__(self, display_item: DisplayItem.DisplayItem, unit_id_model: Model.PropertyModel[str]) -> None:
        super().__init__()
        self.__display_item = display_item
        self.__unit_id_model = unit_id_model
        display_size = calculate_display_size_in_pixels(display_item)
        self.__aspect_ratio = (display_size.width / display_size.height if display_size.height > 0 else 1.0) if display_size.width > 0 else 1.0
        self.__float_to_string_converter = Converter.FloatToStringConverter()
        self.__primary_field = 'width'
        unit_description = pixel_unit_description
        for unit_description_ in svg_export_unit_descriptions:
            if unit_description_.unit_id == unit_id_model.value:
                unit_description = unit_description_
                break
        min_pixels = inch_unit_description.convert_value_to_pixels(3.0)
        max_pixels = inch_unit_description.convert_value_to_pixels(12.0)
        self.__last_quantity = get_quantity_from_pixels(max(min_pixels, min(max_pixels, display_size.width)), unit_description)

    @property
    def title(self) -> typing.Optional[str]:
        """Return the display item title."""
        title_str = self.__display_item.displayed_title
        return title_str

    @property
    def shape_str(self) -> typing.Optional[str]:
        """Return the pixel shape as a string."""
        shape_str = "(" + self.__display_item.data_info.data_shape_str + ")"
        return shape_str

    @property
    def calibrated_shape_str(self) -> typing.Optional[str]:
        """Return the calculated shape as a string."""
        if self.__display_item.data_info.calibrated_dimensional_calibrations_str:
            calibration_str = "(" + self.__display_item.data_info.calibrated_dimensional_calibrations_str + ")"
            return calibration_str
        else:
            return ""

    @property
    def _width(self) -> float:
        """Return the width in pixels.

        If the primary field is width, return the last quantity value directly.

        Otherwise, calculated the width from the last quantity value and the aspect ratio.

        If the last quantity is in pixels, round the value.
        """
        if self.__last_quantity.is_pixels:
            if self.__primary_field == 'width':
                return round(self.__last_quantity.value)
            else:
                return round(self.__last_quantity.value * self.__aspect_ratio)
        else:
            if self.__primary_field == 'width':
                return self.__last_quantity.value
            else:
                return self.__last_quantity.value * self.__aspect_ratio

    @property
    def _height(self) -> float:
        """Return the height in pixels.

        If the primary field is height, return the last quantity value directly.

        Otherwise, calculated the height from the last quantity value and the aspect ratio.

        If the last quantity is in pixels, round the value.
        """
        if self.__last_quantity.is_pixels:
            if self.__primary_field == 'height':
                return round(self.__last_quantity.value)
            else:
                return round(self.__last_quantity.value / self.__aspect_ratio)
        else:
            if self.__primary_field == 'height':
                return self.__last_quantity.value
            else:
                return self.__last_quantity.value / self.__aspect_ratio

    @property
    def width_text(self) -> typing.Optional[str]:
        """Return the width as a string if the primary field is width, otherwise None."""
        if self.__primary_field == 'width':
            return self.__float_to_string_converter.convert(self._width)
        return None

    @width_text.setter
    def width_text(self, value_str: typing.Optional[str]) -> None:
        """Sets the width from a string and sets the primary field to width.

        Also updates the last quantity and signals that all the text fields have changed.
        """
        if value_str:
            value = self.__float_to_string_converter.convert_back(value_str) or 0.0
            self.__last_quantity = self.__last_quantity.with_value(value)
            self.__primary_field = 'width'
        self.notify_property_changed("width_text")
        self.notify_property_changed("height_text")
        self.notify_property_changed("placeholder_width_text")
        self.notify_property_changed("placeholder_height_text")

    @property
    def placeholder_width_text(self) -> str | None:
        """Return the width as a string."""
        return self.__float_to_string_converter.convert(self._width)

    @property
    def height_text(self) -> typing.Optional[str]:
        """Return the height as a string if the primary field is height, otherwise None."""
        if self.__primary_field == 'height':
            return self.__float_to_string_converter.convert(self._height)
        return None

    @height_text.setter
    def height_text(self, value_str: typing.Optional[str]) -> None:
        """Sets the height from a string and sets the primary field to height.

        Also updates the last quantity and signals that all the text fields have changed.
        """
        if value_str:
            value = self.__float_to_string_converter.convert_back(value_str) or 0.0
            self.__last_quantity = self.__last_quantity.with_value(value)
            self.__primary_field = 'height'
        self.notify_property_changed("width_text")
        self.notify_property_changed("height_text")
        self.notify_property_changed("placeholder_width_text")
        self.notify_property_changed("placeholder_height_text")

    @property
    def placeholder_height_text(self) -> str | None:
        """Return the height as a string."""
        return self.__float_to_string_converter.convert(self._height)

    @property
    def unit_description(self) -> UnitDescription:
        """Return the unit description for the current unit id."""
        for unit_description in svg_export_unit_descriptions:
            if unit_description.unit_id == self.__unit_id_model.value:
                return unit_description
        return pixel_unit_description

    @property
    def unit_index(self) -> int:
        """Return the unit index for the current unit description."""
        return svg_export_unit_descriptions.index(self.unit_description)

    @unit_index.setter
    def unit_index(self, unit_index: int) -> None:
        """Set the unit index.

        Saves the unit index to the unit_id_model.

        Then converts the current quantity to the new unit description.

        Then signals that all the text fields have changed.
        """
        if 0 <= unit_index < len(svg_export_unit_descriptions):
            self.__unit_id_model.value = svg_export_unit_descriptions[unit_index].unit_id
        else:
            self.__unit_id_model.value = pixel_unit_description.unit_id
        self.__last_quantity = self.__last_quantity.with_unit_description(self.unit_description)
        self.notify_property_changed("width_text")
        self.notify_property_changed("height_text")
        self.notify_property_changed("placeholder_width_text")
        self.notify_property_changed("placeholder_height_text")

    @property
    def pixel_shape(self) -> Geometry.IntSize:
        if self.__primary_field == 'width':
            return Geometry.IntSize(h=round(self.__last_quantity.pixels / self.__aspect_ratio), w=round(self.__last_quantity.pixels))
        else:
            return Geometry.IntSize(h=round(self.__last_quantity.pixels), w=round(self.__last_quantity.pixels * self.__aspect_ratio))


class ExportSVGHandler(Declarative.Handler):
    def __init__(self, model: ExportSizeModel, get_font_metrics_fn: typing.Callable[[str, str], UserInterface.FontMetrics]) -> None:
        super().__init__()
        self.model = model
        u = Declarative.DeclarativeUI()

        left_column_width = get_font_metrics_fn("normal", "Shape (calibrated units)").width + 20

        right_column_strings = [model.title, model.shape_str, model.calibrated_shape_str]
        right_column_width = max(get_font_metrics_fn("normal", s or str()).width for s in right_column_strings) + 20
        right_column_width = max(right_column_width, left_column_width)

        self.has_calibrated_shape_str = bool(model.calibrated_shape_str)

        unit_titles = [unit_description.title for unit_description in svg_export_unit_descriptions]
        unit_combo_box_width = max(get_font_metrics_fn("normal", s).width for s in unit_titles) + 72

        self.ui_view = u.create_column(
            u.create_row(
                    u.create_label(text="Title", width=left_column_width),
                    u.create_label(text="@binding(model.title)", width=right_column_width),
                    u.create_stretch(),
                    spacing=8
                ),
            u.create_row(
                u.create_label(text="Data shape (Pixels) ", width=left_column_width),
                u.create_label(text="@binding(model.shape_str)", width=right_column_width),
                u.create_stretch(),
                spacing=8
            ),
            u.create_row(
                u.create_label(text="Shape (calibrated units) ", width=left_column_width),
                u.create_label(text="@binding(model.calibrated_shape_str)", width=right_column_width),
                u.create_stretch(),
                spacing=8,
                visible="@binding(has_calibrated_shape_str)"
            ),
            u.create_row(
                u.create_label(text=_("Width"), width=left_column_width),
                u.create_line_edit(
                    placeholder_text="@binding(model.placeholder_width_text)",
                    text="@binding(model.width_text)",
                    width=100
                ),
                u.create_stretch(),
                spacing=8
            ),
            u.create_row(
                u.create_label(text=_("Height"), width=left_column_width),
                u.create_line_edit(
                    placeholder_text="@binding(model.placeholder_height_text)",
                    text="@binding(model.height_text)",
                    width=100
                ),
                u.create_stretch(),
                spacing=8
            ),
            u.create_row(
                u.create_label(text=_("Units"), width=left_column_width),
                u.create_combo_box(
                    items=unit_titles,
                    current_index="@binding(model.unit_index)",
                    width=unit_combo_box_width
                ),
                u.create_stretch(),
                spacing=8
            ),
            spacing=8,
            margin=12,
        )


class ExportSVGDialog:
    def __init__(self, document_controller: DocumentController.DocumentController,
                 display_item: DisplayItem.DisplayItem) -> None:
        super().__init__()
        self.__document_controller = document_controller
        self.__display_item = display_item
        self.__units_model = UserInterface.StringPersistentModel(
            ui=document_controller.ui,
            storage_key="export_units",
            value=pixel_unit_description.unit_id
        )
        self.__model = ExportSizeModel(display_item, self.__units_model)
        self.__handler = ExportSVGHandler(self.__model, document_controller.ui.get_font_metrics)
        self.__init_ui()

    def __init_ui(self) -> None:
        u = Declarative.DeclarativeUI()
        dialog = typing.cast(Dialog.ActionDialog, Declarative.construct(
            self.__document_controller.ui,
            self.__document_controller,
            u.create_modeless_dialog(
                self.__handler.ui_view, title=_("Export SVG")
            ),
            self.__handler
        ))
        dialog.add_button(_("Cancel"), self.__cancel_clicked)
        dialog.add_button(_("Export"), self.__ok_clicked)
        dialog.show()

    def __ok_clicked(self) -> bool:
        pixel_shape = self.__model.pixel_shape
        ui = self.__document_controller.ui
        filter = "SVG File (*.svg);;All Files (*.*)"
        export_dir = ui.get_persistent_string("export_directory", ui.get_document_location())
        export_dir = os.path.join(export_dir, self.__display_item.displayed_title)
        path, selected_filter, selected_directory = self.__document_controller.get_save_file_path(
            _("Export File"), export_dir, filter, None
        )
        if path and not os.path.splitext(path)[1]:
            path = path + os.path.extsep + "svg"
        if path:
            ui.set_persistent_string("export_directory", selected_directory)
            self.__document_controller.export_svg_file(
                DisplayPanel.DisplayPanelUISettings(ui),
                self.__display_item,
                pixel_shape,
                pathlib.Path(path)
            )
        return True

    def __cancel_clicked(self) -> bool:
        return True


@dataclasses.dataclass
class ExportResult:
    data_item_title: str
    error: typing.Optional[str] = None


class ExportResultDialog(Declarative.Handler):
    def __init__(self, ui: UserInterface.UserInterface, document_controller: DocumentController.DocumentController,
                 exports: typing.Sequence[ExportResult], export_folder: pathlib.Path):
        super().__init__()

        self.ui = ui
        self.__document_controller = document_controller
        self.exports = exports
        self.export_folder = export_folder

        # build the UI
        u = Declarative.DeclarativeUI()
        self._build_ui(u)

        # create the dialog and show it.
        export_result_text = _("Export Results")
        items_text = _("Items")
        title_text = f"{export_result_text} ({len(exports)} {items_text})"
        dialog = typing.cast(Dialog.ActionDialog, Declarative.construct(document_controller.ui, document_controller,
                                                                        u.create_modeless_dialog(self.ui_view,
                                                                                                 title=title_text),
                                                                        self))

        dialog.add_button(_("OK"), self.ok_click)
        dialog.show()

    def open_export_folder(self, widget: Declarative.UIWidget) -> bool:
        if platform.system() == 'Windows':
            subprocess.run(['explorer', self.export_folder])
        elif platform.system() == 'Darwin':
            subprocess.Popen(['open', self.export_folder])
        elif platform.system() == 'linux':
            subprocess.Popen(['xdg-open', self.export_folder])
        return True

    def _build_ui(self, u: Declarative.DeclarativeUI) -> None:
        FILE_FIELD_WIDTH = 320
        STATUS_FIELD_WIDTH = 280
        COLUMN_SPACING = 12

        header_labels = [
            u.create_label(text=_('Data Item'), font='bold', width=FILE_FIELD_WIDTH),
            u.create_label(text=_('Error'), font='bold', width=STATUS_FIELD_WIDTH),
        ]

        file_name_labels = list()
        status_labels = list()

        for export in self.exports:
            if export.error:
                status_text = f"\N{WARNING SIGN} {export.error}"
                color = 'red'
                file_name_labels.append(u.create_label(text=export.data_item_title, tool_tip=export.data_item_title, width=FILE_FIELD_WIDTH))
                status_labels.append(u.create_label(text=status_text, tool_tip=export.error, color=color, width=STATUS_FIELD_WIDTH))

        num_errors = len(file_name_labels)
        total_exports = len(self.exports)
        num_successful = total_exports - num_errors
        info_row = u.create_row(u.create_label(text=f"{num_successful} of {total_exports} successful"))

        header_row = u.create_row(*header_labels, u.create_stretch(), spacing=COLUMN_SPACING)

        scroll_area_width = FILE_FIELD_WIDTH + STATUS_FIELD_WIDTH + COLUMN_SPACING * 2 + 24  # 24 is the estimated width of the scrollbar
        scroll_area_height = min(200, 28 + 28 * len(self.exports))  # 28 is the estimated height of a row

        data_row = u.create_scroll_area(
            u.create_column(
                u.create_row(
                    u.create_column(*file_name_labels, u.create_stretch(), spacing=8),
                    u.create_column(*status_labels, u.create_stretch(), spacing=8),
                    u.create_stretch(),
                    spacing=COLUMN_SPACING
                )
            ),
            min_width=scroll_area_width, min_height=scroll_area_height, max_height=240
        )

        path_title = u.create_label(text=_('Directory:'), font='bold')

        path_directory = u.create_label(text=str(self.export_folder), min_width=280, word_wrap=True, size_policy_horizontal='min-expanding', text_alignment_vertical='top')

        path_goto = u.create_row(u.create_push_button(text='Open Directory', on_clicked='open_export_folder'),
                                 u.create_stretch())

        if num_errors == 0:
            self.ui_view = u.create_column(info_row, path_title, path_directory, path_goto, spacing=8, margin=12)
        else:
            self.ui_view = u.create_column(info_row, path_title, path_directory, path_goto, header_row, data_row, spacing=8,
                                           margin=12)

    def ok_click(self) -> bool:
        return True



import typing
import warnings

from nion.swift import Application
from nion.swift import Facade
from nion.ui import Application as ApplicationUI


def main(args: list[typing.Any], bootstrap_args: dict[str, typing.Any]) -> Application.Application:
    # from nion.swift import Application
    warnings.simplefilter("always", RuntimeWarning)
    Facade.initialize()
    app = Application.Application(ApplicationUI.make_ui(bootstrap_args))
    app.initialize(use_root_dir=False)
    Facade.start_server()
    return app
